"""
The European "Gas Day", as specified by ACER
--------------------------------------------

​​​​​​​​The Gas Day establishes the trading window for exchanging day-ahead
and daily gas products on the EU trading platforms. The Gas Day has
a harmonised start and end from 5.00 to 5.00 UTC the following day
for winter time, and from 4.00 to 4.00 UTC the following day when
daylight saving is applied. The harmonised Gas Day promotes
cross-border trading and underpins capacity bundling.​

Source:
https://documents.acer.europa.eu/en/Gas/Framework%20guidelines_and_network%20codes/Pages/Gas-Day.aspx


How to generate the zone in this file
-------------------------------------

 - Open the CET zonefile
   >>> fp = open('<path-to-python-site-packages>/pytz/zoneinfo/CET', 'rb')

 - Copy and modify the latest build_tzinfo file from pytz:
   https://github.com/stub42/pytz/blob/master/src/pytz/tzfile.py

 - Extract variables "zone", "transitions" and "transition_info".

 - Replace offsets and names.
"""

import datetime

from pytz.tzinfo import DstTzInfo
from pytz import _tzinfo_cache


def _build_europe_gas_day_tzinfo():
    """
    Build the Europe/Gas_Day based on the CET time.

    :raises ValueError: When something is wrong with the CET/CEST definition
    """
    zone = 'Europe/Gas_Day'
    transitions = _get_transitions()
    transition_info_cet = _get_transition_info_cet()

    difference_sec = 3600 * 6
    transition_info_gas_day = []
    for dt1, dt2, name in transition_info_cet:
        sec1 = dt1.seconds - difference_sec
        hours1 = sec1 / (60 * 60)
        gas_dt1 = datetime.timedelta(hours=hours1)
        sec2 = dt2.seconds - difference_sec
        hours2 = sec2 / (60 * 60)
        gas_dt2 = datetime.timedelta(hours=hours2)
        if name == 'CET':
            name = 'CET'
        elif name == 'CEST':
            name = 'CEST'
        else:
            raise ValueError("tz name not CET or CEST")
        transition_info_gas_day.append((gas_dt1, gas_dt2, name))

    gas_day_cls = type('Europe/Gas_Day', (DstTzInfo,), dict(
        zone=zone,
        _utc_transition_times=transitions,
        _transition_info=transition_info_gas_day
    ))

    _tzinfo_cache[zone] = gas_day_cls()
    return _tzinfo_cache[zone]


def _get_transition_info_cet():
    """
    Get all transition info for CET.
    """
    return [
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET'),
        (datetime.timedelta(seconds=7200), datetime.timedelta(seconds=3600), 'CEST'),
        (datetime.timedelta(seconds=3600), datetime.timedelta(0), 'CET')
    ]


def _get_transitions():
    """
    All transitions for the CET time zone.
    """
    return [
        datetime.datetime(1, 1, 1, 0, 0),
        datetime.datetime(1916, 4, 30, 22, 0),
        datetime.datetime(1916, 9, 30, 23, 0),
        datetime.datetime(1917, 4, 16, 1, 0),
        datetime.datetime(1917, 9, 17, 1, 0),
        datetime.datetime(1918, 4, 15, 1, 0),
        datetime.datetime(1918, 9, 16, 1, 0),
        datetime.datetime(1940, 4, 1, 1, 0),
        datetime.datetime(1942, 11, 2, 1, 0),
        datetime.datetime(1943, 3, 29, 1, 0),
        datetime.datetime(1943, 10, 4, 1, 0),
        datetime.datetime(1944, 4, 3, 1, 0),
        datetime.datetime(1944, 10, 2, 1, 0),
        datetime.datetime(1945, 4, 2, 1, 0),
        datetime.datetime(1945, 9, 16, 1, 0),
        datetime.datetime(1977, 4, 3, 1, 0),
        datetime.datetime(1977, 9, 25, 1, 0),
        datetime.datetime(1978, 4, 2, 1, 0),
        datetime.datetime(1978, 10, 1, 1, 0),
        datetime.datetime(1979, 4, 1, 1, 0),
        datetime.datetime(1979, 9, 30, 1, 0),
        datetime.datetime(1980, 4, 6, 1, 0),
        datetime.datetime(1980, 9, 28, 1, 0),
        datetime.datetime(1981, 3, 29, 1, 0),
        datetime.datetime(1981, 9, 27, 1, 0),
        datetime.datetime(1982, 3, 28, 1, 0),
        datetime.datetime(1982, 9, 26, 1, 0),
        datetime.datetime(1983, 3, 27, 1, 0),
        datetime.datetime(1983, 9, 25, 1, 0),
        datetime.datetime(1984, 3, 25, 1, 0),
        datetime.datetime(1984, 9, 30, 1, 0),
        datetime.datetime(1985, 3, 31, 1, 0),
        datetime.datetime(1985, 9, 29, 1, 0),
        datetime.datetime(1986, 3, 30, 1, 0),
        datetime.datetime(1986, 9, 28, 1, 0),
        datetime.datetime(1987, 3, 29, 1, 0),
        datetime.datetime(1987, 9, 27, 1, 0),
        datetime.datetime(1988, 3, 27, 1, 0),
        datetime.datetime(1988, 9, 25, 1, 0),
        datetime.datetime(1989, 3, 26, 1, 0),
        datetime.datetime(1989, 9, 24, 1, 0),
        datetime.datetime(1990, 3, 25, 1, 0),
        datetime.datetime(1990, 9, 30, 1, 0),
        datetime.datetime(1991, 3, 31, 1, 0),
        datetime.datetime(1991, 9, 29, 1, 0),
        datetime.datetime(1992, 3, 29, 1, 0),
        datetime.datetime(1992, 9, 27, 1, 0),
        datetime.datetime(1993, 3, 28, 1, 0),
        datetime.datetime(1993, 9, 26, 1, 0),
        datetime.datetime(1994, 3, 27, 1, 0),
        datetime.datetime(1994, 9, 25, 1, 0),
        datetime.datetime(1995, 3, 26, 1, 0),
        datetime.datetime(1995, 9, 24, 1, 0),
        datetime.datetime(1996, 3, 31, 1, 0),
        datetime.datetime(1996, 10, 27, 1, 0),
        datetime.datetime(1997, 3, 30, 1, 0),
        datetime.datetime(1997, 10, 26, 1, 0),
        datetime.datetime(1998, 3, 29, 1, 0),
        datetime.datetime(1998, 10, 25, 1, 0),
        datetime.datetime(1999, 3, 28, 1, 0),
        datetime.datetime(1999, 10, 31, 1, 0),
        datetime.datetime(2000, 3, 26, 1, 0),
        datetime.datetime(2000, 10, 29, 1, 0),
        datetime.datetime(2001, 3, 25, 1, 0),
        datetime.datetime(2001, 10, 28, 1, 0),
        datetime.datetime(2002, 3, 31, 1, 0),
        datetime.datetime(2002, 10, 27, 1, 0),
        datetime.datetime(2003, 3, 30, 1, 0),
        datetime.datetime(2003, 10, 26, 1, 0),
        datetime.datetime(2004, 3, 28, 1, 0),
        datetime.datetime(2004, 10, 31, 1, 0),
        datetime.datetime(2005, 3, 27, 1, 0),
        datetime.datetime(2005, 10, 30, 1, 0),
        datetime.datetime(2006, 3, 26, 1, 0),
        datetime.datetime(2006, 10, 29, 1, 0),
        datetime.datetime(2007, 3, 25, 1, 0),
        datetime.datetime(2007, 10, 28, 1, 0),
        datetime.datetime(2008, 3, 30, 1, 0),
        datetime.datetime(2008, 10, 26, 1, 0),
        datetime.datetime(2009, 3, 29, 1, 0),
        datetime.datetime(2009, 10, 25, 1, 0),
        datetime.datetime(2010, 3, 28, 1, 0),
        datetime.datetime(2010, 10, 31, 1, 0),
        datetime.datetime(2011, 3, 27, 1, 0),
        datetime.datetime(2011, 10, 30, 1, 0),
        datetime.datetime(2012, 3, 25, 1, 0),
        datetime.datetime(2012, 10, 28, 1, 0),
        datetime.datetime(2013, 3, 31, 1, 0),
        datetime.datetime(2013, 10, 27, 1, 0),
        datetime.datetime(2014, 3, 30, 1, 0),
        datetime.datetime(2014, 10, 26, 1, 0),
        datetime.datetime(2015, 3, 29, 1, 0),
        datetime.datetime(2015, 10, 25, 1, 0),
        datetime.datetime(2016, 3, 27, 1, 0),
        datetime.datetime(2016, 10, 30, 1, 0),
        datetime.datetime(2017, 3, 26, 1, 0),
        datetime.datetime(2017, 10, 29, 1, 0),
        datetime.datetime(2018, 3, 25, 1, 0),
        datetime.datetime(2018, 10, 28, 1, 0),
        datetime.datetime(2019, 3, 31, 1, 0),
        datetime.datetime(2019, 10, 27, 1, 0),
        datetime.datetime(2020, 3, 29, 1, 0),
        datetime.datetime(2020, 10, 25, 1, 0),
        datetime.datetime(2021, 3, 28, 1, 0),
        datetime.datetime(2021, 10, 31, 1, 0),
        datetime.datetime(2022, 3, 27, 1, 0),
        datetime.datetime(2022, 10, 30, 1, 0),
        datetime.datetime(2023, 3, 26, 1, 0),
        datetime.datetime(2023, 10, 29, 1, 0),
        datetime.datetime(2024, 3, 31, 1, 0),
        datetime.datetime(2024, 10, 27, 1, 0),
        datetime.datetime(2025, 3, 30, 1, 0),
        datetime.datetime(2025, 10, 26, 1, 0),
        datetime.datetime(2026, 3, 29, 1, 0),
        datetime.datetime(2026, 10, 25, 1, 0),
        datetime.datetime(2027, 3, 28, 1, 0),
        datetime.datetime(2027, 10, 31, 1, 0),
        datetime.datetime(2028, 3, 26, 1, 0),
        datetime.datetime(2028, 10, 29, 1, 0),
        datetime.datetime(2029, 3, 25, 1, 0),
        datetime.datetime(2029, 10, 28, 1, 0),
        datetime.datetime(2030, 3, 31, 1, 0),
        datetime.datetime(2030, 10, 27, 1, 0),
        datetime.datetime(2031, 3, 30, 1, 0),
        datetime.datetime(2031, 10, 26, 1, 0),
        datetime.datetime(2032, 3, 28, 1, 0),
        datetime.datetime(2032, 10, 31, 1, 0),
        datetime.datetime(2033, 3, 27, 1, 0),
        datetime.datetime(2033, 10, 30, 1, 0),
        datetime.datetime(2034, 3, 26, 1, 0),
        datetime.datetime(2034, 10, 29, 1, 0),
        datetime.datetime(2035, 3, 25, 1, 0),
        datetime.datetime(2035, 10, 28, 1, 0),
        datetime.datetime(2036, 3, 30, 1, 0),
        datetime.datetime(2036, 10, 26, 1, 0),
        datetime.datetime(2037, 3, 29, 1, 0),
        datetime.datetime(2037, 10, 25, 1, 0)
    ]



from .base import BaseAPI

from ..metadata import CurveType
from ..parser.metadata import parse_instance_list
from ..parser.periodseries import parse_periodseries, parse_periodseries_list


# Tuple of supported values for Curve.curve_type in the period instances API
CURVE_TYPES = (CurveType.INSTANCE_PERIOD,)


class PeriodInstancesAPI(BaseAPI):
    """
    Period-based series instances API operations. Access these operations via
    an instance of the :py:class:`energyquantified.EnergyQuantified` class:

       >>> eq = EnergyQuantified(api_key="aaaa-bbbb-cccc-dddd")
       >>> eq.period_instances.load(curve, begin, end)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def list(
            self,
            curve,
            tags=None,
            exlude_tags=None,
            limit=20,
            issued_at_latest=None,
            issued_at_earliest=None):
        """
        List instances for the curve. Does not lad any period-based series.

        Filter on the instance `tags` or `exclude_tags` (list of values
        allowed), or restrict the instance issue date by setting
        `issued_at_latest` and/or `issued_at_earliest`.

        Limit the maximum number of returned instances by setting `limit`
        anywhere between 1 and 20.

        This operation works for curves with
        ``curve_type = INSTANCE_PERIOD`` only.

        :param curve: The curve or curve name
        :type curve: :py:class:`energyquantified.metadata.Curve`, str
        :param tags: Filter by instance tags, defaults to None
        :type tags: list, str, optional
        :param exlude_tags: Exclude instance tags, defaults to None
        :type exlude_tags: list, str, optional
        :param limit: Number of instances returned, defaults to 25
        :type limit: int, optional
        :param issued_at_latest: Filter by issue date, defaults to None
        :type issued_at_latest: datetime, date, str, optional
        :param issued_at_earliest: Filter by issue date, defaults to None
        :type issued_at_earliest: datetime, date, str, optional
        :return: A list of :py:class:`energyquantified.metadata.Instance`
        :rtype: list
        """
        # Build URL
        safe_curve = self._urlencode_curve_name(curve, curve_types=CURVE_TYPES)
        url = f"/period-instances/{safe_curve}/list/"
        # Parameters
        params = {}
        self._add_str_list(params, "tags", tags)
        self._add_str_list(params, "exclude-tags", exlude_tags)
        self._add_int(params, "limit", limit, min=1, max=20, required=True)
        self._add_datetime(params, "issued-at-latest", issued_at_latest)
        self._add_datetime(params, "issued-at-earliest", issued_at_earliest)
        # HTTP request
        response = self._get(url, params=params)
        return parse_instance_list(response.json())

    def load(
            self,
            curve,
            begin=None,
            end=None,
            tags=None,
            exlude_tags=None,
            limit=3,
            issued_at_latest=None,
            issued_at_earliest=None):
        """
        Load period-based series instances.

        Filter on the instance `tags` or `exclude_tags` (list of values
        allowed), or restrict the instance issue date by setting
        `issued_at_latest` and/or `issued_at_earliest`.

        This operation works for curves with
        ``curve_type = INSTANCE_PERIOD`` only.

        :param curve: The curve or curve name
        :type curve: :py:class:`energyquantified.metadata.Curve`, str
        :param begin: The begin date-time
        :type begin: date, datetime, str, required
        :param end: The end date-time
        :type end: date, datetime, str, required
        :param tags: Filter by instance tags, defaults to None
        :type tags: list, str, optional
        :param exlude_tags: Exclude instance tags, defaults to None
        :type exlude_tags: list, str, optional
        :param limit: Number of instances returned, defaults to 3
        :type limit: int, optional
        :param issued_at_latest: Filter by issue date, defaults to None
        :type issued_at_latest: datetime, date, str, optional
        :param issued_at_earliest: Filter by issue date, defaults to None
        :type issued_at_earliest: datetime, date, str, optional
        :return: List of :py:class:`energyquantified.data.Periodseries` objects
        :rtype: list
        """
        # Build URL
        safe_curve = self._urlencode_curve_name(curve, curve_types=CURVE_TYPES)
        url = f"/period-instances/{safe_curve}/"
        # Parameters
        params = {}
        self._add_datetime(params, "begin", begin, required=True)
        self._add_datetime(params, "end", end, required=True)
        self._add_str_list(params, "tags", tags)
        self._add_str_list(params, "exclude-tags", exlude_tags)
        self._add_int(params, "limit", limit, min=1, max=20, required=True)
        self._add_datetime(params, "issued-at-latest", issued_at_latest)
        self._add_datetime(params, "issued-at-earliest", issued_at_earliest)
        # HTTP request
        response = self._get(url, params=params)
        return parse_periodseries_list(response.json())

    def latest(
            self,
            curve,
            begin=None,
            end=None,
            issued_at_latest=None):
        """
        Get the latest period-based series instance.

        This operation works for curves with
        ``curve_type = INSTANCE_PERIOD`` only.

        :param curve: The curve or curve name
        :type curve: :py:class:`energyquantified.metadata.Curve`, str
        :param begin: The begin date-time
        :type begin: date, datetime, str, required
        :param end: The end date-time
        :type end: date, datetime, str, required
        :param issued_at_latest: The latest issue date for the loaded instance,\
             defaults to None
        :type issued_at_latest: date, datetime, str, optional
        :return: A period-based series
        :rtype: :py:class:`energyquantified.data.Periodseries`
        """
        # Build URL
        safe_curve = self._urlencode_curve_name(curve, curve_types=CURVE_TYPES)
        url = f"/period-instances/{safe_curve}/get/latest/"
        # Parameters
        params = {}
        self._add_datetime(params, "begin", begin, required=True)
        self._add_datetime(params, "end", end, required=True)
        self._add_datetime(params, "issued-at-latest", issued_at_latest)
        # HTTP request
        response = self._get(url, params=params)
        return parse_periodseries(response.json())

    def get(
            self,
            curve,
            begin=None,
            end=None,
            issued=None,
            tag=""):
        """
        Get a specific period-based series instance.

        :param curve: The curve or curve name
        :type curve: :py:class:`energyquantified.metadata.Curve`, str
        :param begin: The begin date-time
        :type begin: date, datetime, str, required
        :param end: The end date-time
        :type end: date, datetime, str, required
        :param issued: The instance issue date
        :type issued: datetime, date, str, required
        :param tag: The instance tag
        :type tag: str, required
        :return: A period-based series
        :rtype: :py:class:`energyquantified.data.Periodseries`
        """
        # Build URL
        safe_curve = self._urlencode_curve_name(curve, curve_types=CURVE_TYPES)
        url = f"/period-instances/{safe_curve}/get/latest/"
        safe_issued = self._urlencode_datetime(issued, "issued")
        if tag:
            safe_tag = self._urlencode_string(tag, "tag", lambda t: t.strip())
            url = f"/period-instances/{safe_curve}/get/{safe_issued}/{safe_tag}/"
        else:
            url = f"/period-instances/{safe_curve}/get/{safe_issued}/"
        # Parameters
        params = {}
        self._add_datetime(params, "begin", begin, required=True)
        self._add_datetime(params, "end", end, required=True)
        # HTTP request
        response = self._get(url, params=params)
        return parse_periodseries(response.json())



import enum

from dateutil.relativedelta import relativedelta

from .helpers import (
    shift,
    truncate_second, truncate_5min, truncate_10min,
    truncate_15min, truncate_30min, truncate_hour, truncate_date,
    truncate_week, truncate_month, truncate_quarter, truncate_year,
    truncate_season
)


# Lookup dict for frequencies by tag
_frequencies_lookup = {}


class Frequency(enum.Enum):
    """
    Enumerator of valid frequencies for Energy Quantified's API. The
    supported frequencies are:

     * ``P1Y`` – Yearly
     * ``SEASON`` – Summer or winter
     * ``P3M`` – Quarterly
     * ``P1M`` – Monthly
     * ``P1W`` – Weekly
     * ``P1D`` – Daily
     * ``PT1H`` – Hourly
     * ``PT30M`` – 30 minutes
     * ``PT15M`` – 15 minutes
     * ``PT10M`` – 10 minutes
     * ``PT5M`` – 5 minutes
     * ``NONE`` – Tick-data (no frequency)
    """

    # Fixed-interval frequencies
    P1Y    = (True, "years", 1, truncate_year, "Y")
    SEASON = (True, "months", 6, truncate_season, "S")
    P3M    = (True, "months", 3, truncate_quarter, "Q")
    P1M    = (True, "months", 1, truncate_month, "M")
    P1W    = (True, "weeks", 1, truncate_week, "W")
    P1D    = (True, "days", 1, truncate_date, "D")
    PT1H   = (True, "hours", 1, truncate_hour, "H")
    PT30M  = (True, "minutes", 30, truncate_30min, "30min")
    PT15M  = (True, "minutes", 15, truncate_15min, "15min")
    PT10M  = (True, "minutes", 10, truncate_10min, "10min")
    PT5M   = (True, "minutes", 5, truncate_5min, "5min")

    # No fixed interval (i.e. tick data)
    NONE  = (False, "microseconds", 1, lambda dt: dt, None)

    def __init__(self, is_iterable=None, field=None, steps=1,
                 truncate_func=None, pretty_name=None):
        self.is_iterable = is_iterable
        self.ordinal = len(self.__class__.__members__) + 1
        self.field = field
        self.steps = steps
        self.delta = relativedelta(**{ field: steps }) if is_iterable else None
        self.truncate_func = truncate_func
        self.pretty_name = pretty_name
        _frequencies_lookup[self.name.lower()] = self

    @classmethod
    def by_tag(cls, tag):
        """
        Look up frequencies by tag.

        :param tag: The tag to look up
        :type tag: str
        :return: A Frequency for this tag
        :rtype: Frequency
        """
        return _frequencies_lookup[tag.lower()]

    @classmethod
    def is_valid_tag(cls, tag):
        """
        Check whether a frequency tag exists or not.

        :param tag: The tag to look up
        :type tag: str
        :return: True if the frequency tag exists, otherwise False
        :rtype: bool
        """
        return tag.lower() in _frequencies_lookup

    def __lt__(self, b): return self.ordinal < b.ordinal
    def __gt__(self, b): return self.ordinal > b.ordinal
    def __le__(self, b): return self.ordinal <= b.ordinal
    def __ge__(self, b): return self.ordinal >= b.ordinal

    @property
    def tag(self):
        """
        The frequency tag (P1D, PT1H, etc.).
        """
        return self.name

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return self.name

    def matches(self, datetime_obj):
        """
        Check if a date-time aligns with this frequency.

        :param datetime_obj: A date-time
        :type datetime_obj: datetime
        :return: True if it falls into this frequency, otherwise False
        :rtype: bool
        """
        return self.truncate_func(datetime_obj) == datetime_obj

    def truncate(self, datetime_obj):
        """
        Snap a date-time down to the nearest instant in this
        frequency.

        :param datetime_obj: A date-time
        :type datetime_obj: datetime
        :return: A datetime snapped down to the nearest instant in this frequency
        :rtype: datetime
        """
        return self.truncate_func(datetime_obj)

    def get_delta(self, num_steps):
        """
        Create a relativedelta of N steps in this frequency. 0 steps
        means no change. 1 step means one tick forward, -1 step means one
        tick backwards.

        :param num_steps: Number of steps forward (positive) or backwards\
                          (negative)
        :type num_steps: int
        :return: A relativedelta (similar to timedelta)
        :rtype: relativedelta
        """
        # Requires an iterable frequency
        assert self.is_iterable, "%s is not iterable" % self
        return relativedelta(**{ self.field: self.steps * num_steps })

    def shift(self, datetime_obj, num_steps):
        """
        Shift a date-time by N steps in this frequency. 0 steps
        means no change. 1 step means one tick forward, -1 step means one
        tick backwards.

        :param datetime_obj: A date-time
        :type datetime_obj: datetime
        :param num_steps: Number of steps forward (positive) or backwards\
                          (negative)
        :type num_steps: int
        :return: A shifted date-time
        :rtype: datetime
        """
        # Requires an iterable frequency
        assert self.is_iterable, "%s is not iterable" % self
        return shift(datetime_obj, **{ self.field: self.steps * num_steps })

    def between(self, datetime_obj1, datetime_obj2):
        """
        Count number of steps in this frequency between two dates.

        :param datetime_obj1: The first date-time
        :type datetime_obj1: datetime
        :param datetime_obj2: The second date-time
        :type datetime_obj2: datetime
        :return: The number of time this frequency fits between the two dates
        :rtype: int
        """
        # Requires an iterable frequency
        assert self.is_iterable, "%s is not iterable" % self
        # Check that they are in the frequency
        assert self.matches(datetime_obj1), "First date must match frequency"
        assert self.matches(datetime_obj2), "Second date must match frequency"
        d1, d2 = datetime_obj1, datetime_obj2
        # Same date
        if d1 == d2:
            return 0
        # Initiate counters and deltas
        delta = self.get_delta(1)
        d = d1
        i = 0
        # Forwards or backwards
        if d1 < d2:
            while d < d2:
                d = d + delta
                i = i + 1
        else:
            while d > d2:
                d = d - delta
                i = i - 1
        # The conclusion
        return i


# Additional lookups for

_frequencies_lookup["pt60m"] = Frequency.PT1H


# Constants

NONE   = Frequency.NONE

PT5M   = Frequency.PT5M
PT10M  = Frequency.PT10M
PT15M  = Frequency.PT15M
PT30M  = Frequency.PT30M
PT60M  = Frequency.PT1H
PT1H   = Frequency.PT1H
P1D    = Frequency.P1D
P1W    = Frequency.P1W
P1M    = Frequency.P1M
P3M    = Frequency.P3M
SEASON = Frequency.SEASON
P1Y    = Frequency.P1Y


# Lookup

LOOKUP = { str(f): f for f in list(Frequency) }
LOOKUP["PT60M"] = PT1H



import sys

from ..exceptions import PageError


class Page(list):
    """
    A Page is a subclass of the standard `list` type in Python
    with additional attributes for paging.

    It also has methods for loading the next/previous page, and to
    print the page to stdout.
    """

    def __init__(
            self,
            elements,
            page=None,
            page_size=None,
            total_items=None,
            first_page=None,
            last_page=None,
            page_load_func=None,
        ):
        super().__init__(elements)
        # --- Public members ---
        #: The current page
        self.page = page
        #: The page size
        self.page_size = page_size
        #: Total number of items found in search across all pages
        self.total_items = total_items
        #: The first page (is always 1)
        self.first_page = first_page
        #: The last page in the search
        self.last_page = last_page
        #: The total number of pages
        self.total_pages = last_page
        self._page_load_func = page_load_func

    def append(self, value):
        raise NotImplementedError("Page does not support append")

    def extend(self, iterable):
        raise NotImplementedError("Page does not support extend")

    def insert(self, index, value):
        raise NotImplementedError("Page does not support insert")

    def remove(self, value):
        raise NotImplementedError("Page does not support remove")

    def pop(self, index=-1):
        raise NotImplementedError("Page does not support pop")

    def clear(self):
        raise NotImplementedError("Page does not support clear")

    def __setitem__(self, key, value):
        raise NotImplementedError("Page does not support __setitem__")

    def __delitem__(self, key):
        raise NotImplementedError("Page does not support __delitem__")

    def __add__(self, rhs):
        raise NotImplementedError("Page does not support __add__")

    def __iadd__(self, rhs):
        raise NotImplementedError("Page does not support __iadd__")

    def __mul__(self, rhs):
        raise NotImplementedError("Page does not support multiply")

    def __rmul__(self, rhs):
        raise NotImplementedError("Page does not support multiply")

    def __imul__(self, rhs):
        raise NotImplementedError("Page does not support multiply")

    def has_next_page(self):
        """
        Check if there is a next page.

        :return: True when there is a next page
        :rtype: bool
        """
        return self.page is not None and self.page < self.last_page

    def has_previous_page(self):
        """
        Check if there is a previous page.

        :return: True when there is a previous page
        :rtype: bool
        """
        return self.page is not None and self.page > self.first_page

    def get_next_page(self):
        """
        Get the next page. Will perform an HTTP request if data is not
        already in local cache.

        :raises PageError: When there is no next page
        :raises PageError: There is no support for loading the next page
        :return: The next page
        :rtype: Page
        """
        if not self.has_next_page():
            raise PageError("No more pages available")
        if not self._page_load_func:
            raise PageError("Cannot load more pages")
        return self._page_load_func(page=self.page + 1)

    def get_previous_page(self):
        """
        Get the previous page. Will perform an HTTP request if data is not
        already in local cache.

        :raises PageError: When there is no previous page
        :raises PageError: There is no support for loading the previous page
        :return: The previous page
        :rtype: Page
        """
        if not self.has_previous_page():
            raise PageError("No previous page available")
        if not self._page_load_func:
            raise PageError("Cannot load more pages")
        return self._page_load_func(page=self.page - 1)

    def print(self, file=sys.stdout):
        """
        Print the page with all its attributes.

        :param file: A file descriptor, defaults to sys.stdout
        :type file: file descriptor, optional
        """
        print(
            f"Page:\n"
            f"   current = {self.page}\n"
            f"   page-size = {self.page_size}\n"
            f"   total-items = {self.total_items}\n"
            f"   total-pages = {self.total_pages}\n"
            f"   items:\n",
            file=file
        )
        for item in self:
            print("    –", item, file=file)

    @staticmethod
    def _response_to_page(items, response, load_func=None):
        """
        Private utility method to convert an HTTP response and a list of
        items to a Page object.
        """
        # Parse pagination from response headers
        headers = response.headers
        page = int(headers.get("X-Current-Page"))
        page_size = int(headers.get("X-Page-Size"))
        first_page = int(headers.get("X-First-Page"))
        last_page = int(headers.get("X-Last-Page"))
        total_items = int(headers.get("X-Total-Items"))
        # To a page instance
        return Page(
            items,
            page=page,
            page_size=page_size,
            first_page=first_page,
            last_page=last_page,
            total_items=total_items,
            page_load_func=load_func
        )



"""
Energy Quantified Time Series API client
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

# Add a default void log handler to avoid "No handler found" warnings
import logging
from logging import NullHandler

logging.getLogger(__name__).addHandler(NullHandler())


# Package info
from .__version__ import (
    __title__,
    __description__,
    __url__,
    __version__,
)


# The client implementation
from .base import EnergyQuantified, RealtoConnection


__all__ = [
    "EnergyQuantified",
    "RealtoConnection",
]



from .base import BaseAPI

from ..exceptions import ValidationError
from ..metadata import CurveType, OHLCField
from ..parser.ohlc import parse_ohlc_response
from ..parser.timeseries import parse_timeseries
from ..parser.periodseries import parse_periodseries


# Tuple of supported values for Curve.curve_type in the time series API
CURVE_TYPES = (CurveType.OHLC,)

class OhlcAPI(BaseAPI):
    """
    Operations for OHLC data retrieval. Access these operations via an
    instance of the :py:class:`energyquantified.EnergyQuantified` class:

       >>> eq = EnergyQuantified(api_key="aaaa-bbbb-cccc-dddd")
       >>> eq.ohlc.load(curve, begin, end)
    """

    def load(
            self,
            curve,
            begin=None,
            end=None,
            period=None,
            delivery=None,
            front=None):
        """
        Load OHLC data for a :py:class:`energyquantified.metadata.Curve`.

        This operation works for curves with ``curve_type = OHLC`` only.

        :param curve: The curve or curve name
        :type curve: :py:class:`energyquantified.metadata.Curve`, str
        :param begin: The begin date
        :type begin: date, str, required
        :param end: The end date
        :type end: date, str, required
        :param period: Filter on contract period (day, week, month etc.),\
            defaults to None
        :type period: ContractPeriod, str, optional
        :param delivery: Filter on delivery date, requires parameter ``period``\
            to be set; cannot be used together with ``front``, defaults to None
        :type delivery: date, str, optional
        :param front: Filter on front contract, requires parameter ``period``\
            to be set; cannot be used together with ``delivery``,\
            defaults to None
        :type front: int, optional
        :return: A list of OHLC objects
        :rtype: :py:class:`energyquantified.data.OHLCList`
        """
        # Build URL
        safe_curve = self._urlencode_curve_name(curve, curve_types=CURVE_TYPES)
        url = f"/ohlc/{safe_curve}/"
        # Parameters
        params = {}
        self._add_date(params, "begin", begin, required=True)
        self._add_date(params, "end", end, required=True)
        self._add_contract_period(params, "period", period)
        self._add_int(params, "front", front, min=1)
        self._add_date(params, "delivery", delivery)
        # Additional checks
        has_period = "period" in params
        if has_period:
            if "front" in params and "delivery" in params:
                raise ValidationError(reason=(
                    "At most one of the following fields may be set: "
                    "'front', 'delivery'"
                ))
        else:
            if "front" in params:
                raise ValidationError(reason=(
                    "Parameter 'front' can only be used in together with "
                    "'period'"
                ))
            if "delivery" in params:
                raise ValidationError(reason=(
                    "Parameter 'delivery' can only be used in together with "
                    "'period'"
                ))
        # HTTP request
        response = self._get(url, params=params)
        return parse_ohlc_response(response.json())

    def latest(
            self,
            curve,
            date=None):
        """
        Select all OHLC for specific trading day, while defaulting to the
        latest available OHLC data.

        If ``date`` is given, this method will try to fetch OHLC data for
        that trading day. When there is no data for the given day, OHLC data
        will be loaded for the closest trading day earlier in time with data.

        This operation works for curves with ``curve_type = OHLC`` only.

        :param curve: The curve or curve name
        :type curve: :py:class:`energyquantified.metadata.Curve`, str
        :param date: The trading date, defaults to today
        :type date: date, str, required
        :return: A list of OHLC objects
        :rtype: :py:class:`energyquantified.data.OHLCList`
        """
        # Build URL
        safe_curve = self._urlencode_curve_name(curve, curve_types=CURVE_TYPES)
        url = f"/ohlc/{safe_curve}/latest/"
        # Parameters
        params = {}
        self._add_date(params, "date", date)
        # HTTP request
        response = self._get(url, params=params)
        return parse_ohlc_response(response.json())

    def load_delivery_as_timeseries(
            self,
            curve,
            begin=None,
            end=None,
            period=None,
            delivery=None,
            field=OHLCField.SETTLEMENT,
            fill=None):
        """
        Load historical OHLC data for specific contract, and convert it to a
        :py:class:`energyquantified.data.Timeseries`.

        Defaults to use the settlement field, but you can select any field
        you want. All other parameters are required.

        This operation works for curves with ``curve_type = OHLC`` only.

        :param curve: The curve or curve name
        :type curve: :py:class:`energyquantified.metadata.Curve`, str
        :param begin: The begin date
        :type begin: date, str, required
        :param end: The end date
        :type end: date, str, required
        :param period: The contract period (week, month, quarter)
        :type period: ContractPeriod, str, required
        :param delivery: The delivery date for the contract
        :type delivery: date, str, required
        :param field: The field (close, settlement, etc.) to extract to a\
            time series, defaults to ``OHLCField.SETTLEMENT``
        :type field: OHLCField, str, required
        :param fill: How to handle days without trades. Allowed values are:\
            ``no-fill`` do nothing, ``fill-holes`` fill in holes with data\
            from previous trading day, ``forward-fill`` fill in all blanks\
            with data from the previous trading day (also into the future).\
            Defaults to ``no-fill``.
        :type fill: str, optional
        :return: A time series
        :rtype: :py:class:`energyquantified.data.Timeseries`
        """
        # Curve name for URL
        safe_curve = self._urlencode_curve_name(curve, curve_types=CURVE_TYPES)
        # Parameters
        params = {}
        self._add_date(params, "begin", begin, required=True)
        self._add_date(params, "end", end, required=True)
        self._add_contract_period(params, "period", period, required=True)
        self._add_date(params, "delivery", delivery, required=True)
        self._add_fill(params, "fill", fill)
        # Build URL
        field = self._urlencode_ohlc_field(field, "field")
        url = f"/ohlc/{safe_curve}/timeseries/{field}/"
        # HTTP request
        response = self._get(url, params=params)
        return parse_timeseries(response.json())

    def load_front_as_timeseries(
            self,
            curve,
            begin=None,
            end=None,
            period=None,
            front=None,
            field=OHLCField.SETTLEMENT,
            fill=None):
        """
        Load historical OHLC data for a continuous front contract, and convert
        it to a :py:class:`energyquantified.data.Timeseries`.

        Defaults to use the settlement field, but you can select any field
        you want. All other parameters are required.

        This operation works for curves with ``curve_type = OHLC`` only.

        :param curve: The curve or curve name
        :type curve: :py:class:`energyquantified.metadata.Curve`, str
        :param begin: The begin date
        :type begin: date, str, required
        :param end: The end date
        :type end: date, str, required
        :param period: The contract period (week, month, quarter)
        :type period: ContractPeriod, str, required
        :param front: The front contract (1=front, 2=second front, etc.)
        :type front: int, required
        :param field: The field (close, settlement, etc.) to extract to a\
            time series, defaults to ``OHLCField.SETTLEMENT``
        :type field: OHLCField, str, required
        :param fill: How to handle days without trades. Allowed values are:\
            ``no-fill`` do nothing, ``fill-holes`` fill in holes with data\
            from previous trading day, ``forward-fill`` fill in all blanks\
            with data from the previous trading day (also into the future).\
            Defaults to ``no-fill``.
        :type fill: str, optional
        :return: A time series
        :rtype: :py:class:`energyquantified.data.Timeseries`
        """
        # Curve name for URL
        safe_curve = self._urlencode_curve_name(curve, curve_types=CURVE_TYPES)
        # Parameters
        params = {}
        self._add_date(params, "begin", begin, required=True)
        self._add_date(params, "end", end, required=True)
        self._add_contract_period(params, "period", period, required=True)
        self._add_int(params, "front", front, min=1, required=True)
        self._add_fill(params, "fill", fill)
        # Build URL
        field = self._urlencode_ohlc_field(field, "field")
        url = f"/ohlc/{safe_curve}/timeseries/{field}/"
        # HTTP request
        response = self._get(url, params=params)
        return parse_timeseries(response.json())

    def latest_as_periods(
            self,
            curve,
            field=OHLCField.SETTLEMENT,
            date=None):
        """
        Load all OHLC rows from a single trading day, sort them, and
        merge/convert them to a continuous series.

        It defaults to using the latest prices available, therefore
        "latest_as_period".

        If ``date`` is given, this method will try to fetch OHLC data for
        that trading day. When there is no data for the given day, OHLC data
        will be loaded for the closest trading day earlier in time with data.

        By default, this method uses the settlement price. Select another
        field, such as close price, by setting the `field` parameter.

        This operation works for curves with ``curve_type = OHLC`` only.

        :param curve: The curve or curve name
        :type curve: :py:class:`energyquantified.metadata.Curve`, str
        :param field: The field to generate the series from, \
            defaults to OHLCField.SETTLEMENT
        :type field: OHLCField, str, optional
        :param date: The trading date, defaults to today
        :type date: date, str, required
        :return: A period-based series
        :rtype: :py:class:`energyquantified.data.Periodseries`
        """
        # Build URL
        safe_curve = self._urlencode_curve_name(curve, curve_types=CURVE_TYPES)
        # Parameters
        params = {}
        self._add_date(params, "date", date)
        # Build URL
        field = self._urlencode_ohlc_field(field, "field")
        url = f"/ohlc/{safe_curve}/latest/periods/{field}/"
        # HTTP request
        response = self._get(url, params=params)
        return parse_periodseries(response.json())



from ..data import SRMC, SRMCOptions, OHLCList
from ..exceptions import ParseException
from ..metadata import Area
from .ohlc import parse_ohlc_list
from .timeseries import parse_timeseries
from .periodseries import parse_periodseries
from .metadata import parse_curve, parse_contract


def parse_srmc_response(json):
    # Get curve
    curve = json.get('curve')
    if curve is not None:
        curve = parse_curve(curve)
    # Get contract
    contract = json.get('contract')
    if contract is not None:
        contract = parse_contract(contract)
    # Get SRMC options
    options = json.get('srmc_options')
    if options is not None:
        options = parse_srmc_options(options)
    # Check for OHLC response
    data = json.get('ohlc')
    if data:
        ohlc = parse_ohlc_list(data)
        return SRMC(
            curve=curve,
            contract=contract,
            options=options,
            ohlc=OHLCList(ohlc)
        )
    # Check for timeseries response
    data = json.get('timeseries')
    if data:
        timeseries = parse_timeseries(data)
        return SRMC(
            curve=curve,
            contract=contract,
            options=options,
            timeseries=timeseries
        )
    # Check for period-based series response
    data = json.get('periodseries')
    if data:
        periodseries = parse_periodseries(data)
        return SRMC(
            curve=curve,
            contract=contract,
            options=options,
            periodseries=periodseries
        )
    # Something went wrong
    raise ParseException(
        "Failed to parse SRMC response, expected either ohlc data, "
        "timeseries or periodseries, but got neither"
    )

def parse_srmc_options(opts):
    # Check if there are any srmc options
    if not opts:
        return None
    # Get options
    fuel_type = opts.get('fuel_type')
    api2_tonne_to_mwh = opts.get('api2_tonne_to_mwh')
    gas_therm_to_mwh = opts.get('gas_therm_to_mwh')
    efficiency = opts.get('efficiency')
    carbon_emissions = opts.get('carbon_emissions')
    carbon_tax_area = opts.get('carbon_tax_area')
    if carbon_tax_area and Area.is_valid_tag(carbon_tax_area):
        carbon_tax_area = Area.by_tag(carbon_tax_area)
    else:
        carbon_tax_area = None
    # Create srmc options object
    srmc_options = SRMCOptions(
        fuel_type=fuel_type,
        api2_tonne_to_mwh=api2_tonne_to_mwh,
        gas_therm_to_mwh=gas_therm_to_mwh,
        efficiency=efficiency,
        carbon_emissions=carbon_emissions,
        carbon_tax_area=carbon_tax_area
    )
    # SRMC options
    return srmc_options



import logging
import time
import threading

from ..time import now


log = logging.getLogger(__name__)


class RateLimiter:
    """
    A rate limit that sleeps until a new call is allowed. Rate limiting
    will ensure that there is always a minimum delay between each call.

    Defaults to a delay of 1.0 second per call. The rate limiter is thread-safe.

    Example usage:

        # Create a rate limit with a delay of 2.5 second per call
        rate_limiter = RateLimit(rate_per_second=2.5)

        # Rate limit
        for foo in bar:
            rate_limiter()
            ...
    """

    def __init__(self, delay=1.0):
        assert delay is not None, "delay cannot be None"
        self.delay = delay
        self._previous = None
        self._lock = threading.RLock()

    def __call__(self):
        """
        Waits and returns when
        """
        if not self.delay:
            return # Dont impose delay
        else:
            self._lock.acquire()
            try:
                self._rate_limited()
            finally:
                self._lock.release()

    def _rate_limited(self):
        """
        Perform rate limiting and let thread sleep if necessary.
        """
        # Check if we must sleep, except for the first time
        if self._previous:
            diff = (now() - self._previous).total_seconds()
            if diff < self.delay:
                wait_for = (self.delay - diff) + 0.0001
                self._sleep(wait_for)
        # Set previous pass-through time end time
        self._previous = now()

    def _sleep(self, seconds):
        log.debug("Sleep for %.2fs" % seconds)
        time.sleep(seconds)



from dateutil import parser

from ..data import (
    Timeseries,
    TimeseriesList,
    Value,
    ScenariosValue,
    MeanScenariosValue
)
from ..exceptions import ParseException
from ..time import to_timezone
from .metadata import (
    parse_curve,
    parse_instance,
    parse_resolution,
    parse_contract
)


def parse_timeseries_list(json):
    """
    Parse a JSON list-response from the API a list of Timeseries objects with
    Curve, Resolution, Instance and Contract.
    """
    if not isinstance(json, list):
        raise ParseException(
            f"Expected list of time series JSON objects, found: {type(json)}"
        )
    return TimeseriesList(parse_timeseries(json_obj) for json_obj in json)


def parse_timeseries(json):
    """
    Parse a JSON response from the API into a Timeseries object with
    Curve, Resolution, Instance and Contract.
    """
    # Parse the Resolution
    resolution = json.get("resolution")
    if resolution:
        resolution = parse_resolution(resolution)
    # Parse the Curve
    curve = json.get("curve")
    if curve:
        curve = parse_curve(curve)
    # Parse the instance
    instance = json.get("instance")
    if instance:
        instance = parse_instance(instance, curve=curve)
    # Parse the contract
    contract = json.get("contract")
    if contract:
        contract = parse_contract(contract)
    # Parse scenario names
    scenario_names = json.get("scenario_names") or None
    # Parse data
    data = _parse_data(json.get("data"), resolution=resolution)

    return Timeseries(
        curve=curve,
        resolution=resolution,
        instance=instance,
        contract=contract,
        scenario_names=scenario_names,
        data=data
    )


def _parse_data(json, resolution=None):
    """
    Parse the data[] part of the json response
    """
    # Empty time series
    if not json:
        return []
    # Look at the first element
    element = json[0]
    has_value = "v" in element
    has_scenarios = "s" in element
    # Regular (datetime, value) time series
    if has_value and not has_scenarios:
        return _parse_value_data(json, resolution=resolution)
    # Scenario (datetime, scenarios[]) time series
    if not has_value and has_scenarios:
        return _parse_scenarios_data(json, resolution=resolution)
    # Mean + scenario (datetime, mean value, scenarios[]) time series
    if has_value and has_scenarios:
        return _parse_mean_scenarios_data(json, resolution=resolution)
    # Well, this is awkward...
    raise ValueError("Timeseries data is missing both value and scenarios")


def _parse_value_data(json, resolution=None):
    timezone = resolution.timezone

    data = []
    for element in json:
        datetime_obj = to_timezone(
            parser.isoparse(element.get("d")),
            tz=timezone
        )
        value = element.get("v")
        data.append(Value(datetime_obj, value))

    return data


def _parse_scenarios_data(json, resolution=None):
    timezone = resolution.timezone

    data = []
    for element in json:
        datetime_obj = to_timezone(
            parser.isoparse(element.get("d")),
            tz=timezone
        )
        scenarios = tuple(element.get("s"))
        data.append(ScenariosValue(datetime_obj, scenarios))

    return data


def _parse_mean_scenarios_data(json, resolution=None):
    timezone = resolution.timezone

    data = []
    for element in json:
        datetime_obj = to_timezone(
            parser.isoparse(element.get("d")),
            tz=timezone
        )
        mean = element.get("v")
        scenarios = tuple(element.get("s"))
        data.append(MeanScenariosValue(datetime_obj, mean, scenarios))

    return data


import logging
import time

import requests

from ..__version__ import __version__

from .rate_limiter import RateLimiter
from .retry import Retry


# Identify the API client
USER_AGENT = f"Python/EQ-API-Client {__version__}"

# Suppress SSL certificate warnings
# TODO This should probably not be handled here
requests.packages.urllib3.disable_warnings()

# Logging
log = logging.getLogger(__name__)


class Session:
    """
    An HTTP session object wrapper around requests. Enhancements:

     * Configurable rate limiting
     * Configurable retries on server errors
    """

    def __init__(self, timeout=30, max_redirects=5, verify=True, headers={},
            auth=None, base_url=None, delay=None, retry=None):
        # Rate limit and retry
        self.rate_limiter = RateLimiter(delay) if delay else RateLimiter()
        self.retry = retry if retry else Retry()
        # Default arguments
        self.verify = verify
        self.timeout = timeout
        self.max_redirects = max_redirects
        self.headers = headers
        self.auth = auth
        self.base_url = base_url
        # Requests
        self._requests = 0
        # Session
        self.session = requests.Session()
        self.session.max_redirects = max_redirects
        self.session.headers.update({"User-agent": USER_AGENT})
        self.session.headers.update(self.headers)
        self.session.hooks["response"] = [self._request_inc]

    def _request_inc(self, response, *args, **kwargs):
        self._requests += 1

    @property
    def requests(self):
        return self._requests

    def _prepare_args(self, **kwargs):
        return {
            "timeout": kwargs.get("timeout", self.timeout),
            "verify": kwargs.get("verify", self.verify),
            "auth": kwargs.get("auth", self.auth),
            "headers": kwargs.get("headers"),
            "params": kwargs.get("params"),
            "data": kwargs.get("data"),
            "json": kwargs.get("json"),
        }

    def get(self, url, retry=True, **kwargs):
        """
        Perform a HTTP GET request.
        """
        return self.request("get", url, retry=retry, **kwargs)

    def post(self, url, retry=False, **kwargs):
        """
        Perform a HTTP POST request.
        """
        return self.request("post", url, retry=retry, **kwargs)

    def request(self, method, url, retry=False, **kwargs):
        """
        Perform a request using a HTTP verb of your choice.
        """
        # Prepare arguments
        args = self._prepare_args(**kwargs)

        # Append base url
        if self.base_url:
            url = f"{self.base_url}{url}"

        # Make the function that does the work
        def _func():
            self.rate_limiter()
            log.debug(
                "HTTP %s %s %s"
                % (method.upper(), url, args.get("params") or "")
            )
            return self.session.request(method, url, **args)

        # Attach retry logic, if any
        func = self.retry(_func) if retry and self.retry else _func

        # Perform request
        return func()



from collections import namedtuple
from datetime import datetime
import enum
import numbers
import sys


from ..time import Resolution
from ..utils.pandas import timeseries_to_dataframe, timeseries_list_to_dataframe
from .base import Series


class ValueType(enum.Enum):
    """
    Enumerator of supported value types for ``timeseries.data[]``.
    """
    #: Single value. See :py:class:`energyquantified.data.Value`.
    VALUE = "VALUE"
    #: Only scenarios. See :py:class:`energyquantified.data.ScenariosValue`.
    SCENARIOS = "SCENARIOS"
    #: A value and scenarios. See
    #: :py:class:`energyquantified.data.MeanScenariosValue`.
    MEAN_AND_SCENARIOS = "MEAN_AND_SCENARIOS"


class Value(namedtuple("Value", ("date", "value"))):
    """
    A time series value.

    Implemented as a namedtuple of (date, value).

    .. py:attribute:: date

        The date-time for a time series value (index)

    .. py:attribute:: value

        The numeric value
    """

    def value_type(self):
        """
        Get the value type for this tuple.

        :return: The value type for this tuple
        :rtype: ValueType
        """
        return ValueType.VALUE

    def validate(self):
        """
        Check if this time series value is in a valid format.

        :raises ValueError: When this value tuple is invalid
        """
        if not isinstance(self.date, datetime):
            raise ValueError("Value.date is not a datetime")
        if not (self.value is None or isinstance(self.value, (numbers.Number))):
            raise ValueError("Value.value must either be a number or None")

    def num_scenarios(self):
        """
        Count number of scenarios in this data point.

        :return: Always 0, as Value-types has no scenarios
        :rtype: int
        """
        return 0

    def has_scenarios(self):
        """
        Check whether or not this data point has `support` for scenarios.

        :return: Always False, as Value-types has no scenarios
        :rtype: bool
        """
        return False

    def print(self, file=sys.stdout):
        """
        Print this time series value.
        """
        dt = self.date.isoformat(sep=' ')
        v = self.value
        print(f"{dt}\t{v:13.2f}", file=file)

    def to_dict(self):
        """
        Convert a time series value to a dict in the same short
        format as the API response.

        :return: A dict of this value in the same format as the API response
        :rtype: dict
        """
        return {
            "d": self.date.isoformat(sep=' '),
            "v": self.value
        }

    def __str__(self):
        dt = self.date.isoformat(sep=' ')
        v = self.value
        return f"<Value: date={dt}, value={v}>"

    def __repr__(self):
        return str(self)


class ScenariosValue(namedtuple("Value", ("date", "scenarios"))):
    """
    A time series value of scenarios.

    Implemented as a namedtuple of (date, scenarios).

    .. py:attribute:: date

        The date-time for a time series value (index)

    .. py:attribute:: scenarios

        A tuple of scenario values
    """

    def value_type(self):
        """
        Get the value type for this tuple.

        :return: The value type for this tuple
        :rtype: ValueType
        """
        return ValueType.SCENARIOS

    def validate(self):
        """
        Check if this time series value is in a valid format.

        :raises ValueError: When this value tuple is invalid
        """
        if not isinstance(self.date, datetime):
            raise ValueError("ScenariosValue.date is not a datetime")
        if not (self.scenarios is tuple):
            raise ValueError("ScenariosValue.scenarios is not a tuple")
        if not all(
                (s is None or isinstance(s, (numbers.Number)))
                for s in self.scenarios
            ):
            raise ValueError(
                "ScenariosValue.scenarios has members that aren't numbers"
            )

    def num_scenarios(self):
        """
        Count number of scenarios in this data point.

        :return: Number of scenarios in this data point.
        :rtype: int
        """
        return len(self.scenarios)

    def has_scenarios(self):
        """
        Check whether or not this data point has `support` for scenarios.

        :return: Always True, as ScenariosValue-types has scenarios
        :rtype: bool
        """
        return True

    def print(self, file=sys.stdout):
        """
        Print this time series value.
        """
        dt = self.date.isoformat(sep=' ')
        s = self.scenarios
        print(f"{dt}\t{s}", file=file)

    def to_dict(self):
        """
        Convert a time series value to a dict in the same short
        format as the API response.

        :return: A dict of this value in the same format as the API response
        :rtype: dict
        """
        return {
            "d": self.date.isoformat(sep=' '),
            "s": self.scenarios
        }

    def __str__(self):
        dt = self.date.isoformat(sep=' ')
        s = self.scenarios
        return f"<ScenariosValue: date={dt}, scenarios={s}>"

    def __repr__(self):
        return str(self)


class MeanScenariosValue(namedtuple("Value", ("date", "value", "scenarios"))):
    """
    A time series value with scenarios and a mean value.

    Implemented as a namedtuple of (date, value, scenarios).

    .. py:attribute:: date

        The date-time for a time series value (index)

    .. py:attribute:: value

        The numeric value

    .. py:attribute:: scenarios

        A tuple of scenario values
    """

    def value_type(self):
        """
        Get the value type for this tuple.

        :return: The value type for this tuple
        :rtype: ValueType
        """
        return ValueType.MEAN_AND_SCENARIOS

    def validate(self):
        """
        Check if this time series value is in a valid format.

        :raises ValueError: When this value tuple is invalid
        """
        if not isinstance(self.date, datetime):
            raise ValueError("ScenariosValue.date is not a datetime")
        if not (self.value is None or isinstance(self.value, (numbers.Number))):
            raise ValueError("Value.value must either be a number or None")
        if not (self.scenarios is tuple):
            raise ValueError("ScenariosValue.scenarios is not a tuple")
        if not all(
                (s is None or isinstance(s, (numbers.Number)))
                for s in self.scenarios
            ):
            raise ValueError(
                "MeanScenariosValue.scenarios has members that aren't numbers"
            )

    def num_scenarios(self):
        """
        Count number of scenarios in this data point.

        :return: Number of scenarios in this data point.
        :rtype: int
        """
        return len(self.scenarios)

    def has_scenarios(self):
        """
        Check whether or not this data point has `support` for scenarios.

        :return: Always True, as MeanScenariosValue-types has scenarios
        :rtype: bool
        """
        return True

    def print(self, file=sys.stdout):
        """
        Print this time series value.
        """
        dt = self.date.isoformat(sep=' ')
        v = self.value
        s = self.scenarios
        print(f"{dt}\t{v:13.2f}\t{s}", file=file)

    def to_dict(self):
        """
        Convert a time series value to a dict in the same short
        format as the API response.

        :return: A dict of this value in the same format as the API response
        :rtype: dict
        """
        return {
            "d": self.date.isoformat(sep=' '),
            "v": self.value,
            "s": self.scenarios
        }

    def __str__(self):
        dt = self.date.isoformat(sep=' ')
        v = self.value
        s = self.scenarios
        return f"<MeanScenariosValue: date={dt}, value={v}, scenarios={s}>"

    def __repr__(self):
        return str(self)


class Timeseries(Series):
    """
    A time series with metadata.

    :param curve: The curve, defaults to None
    :type curve: Curve, optional
    :param name: A name which is used as column name when converted to \
        a `pandas.DataFrame`, defaults to None
    :type name: str, optional
    :param resolution: The resolution of the time series, defaults to None
    :type resolution: Resolution, optional
    :param instance: The instance, defaults to None
    :type instance: Instance, optional
    :param data: A list of values (Value, ScenariosValue, or MeanScenariosValue)
    :type data: list[]
    :param scenario_names: A list of scenario names, if any
    :type scenario_names: list[str], optional
    """

    def __init__(self, data=None, scenario_names=None, **kwargs):
        super().__init__(**kwargs)
        assert isinstance(self.resolution, Resolution), (
            "Timeseries.resolution is required"
        )
        self.data = data or []
        self.scenario_names = scenario_names or []
        self._fix_scenario_names()

    def __str__(self):
        items = []
        items.append(f"resolution={self.resolution}")
        items.append(f"curve=\"{self.curve}\"")
        if self.instance:
            items.append(f"instance={self.instance}")
        if self.has_scenarios():
            items.append(f"scenario_names={self.scenario_names}")
        if self.has_data():
            items.append(f"begin=\"{self.begin().isoformat(sep=' ')}\"")
            items.append(f"end=\"{self.end().isoformat(sep=' ')}\"")
        else:
            items.append("EMPTY")
        return f"<Timeseries: {', '.join(items)}>"

    def __repr__(self):
        return str(self)

    def _fix_scenario_names(self):
        if self.data and self.data[0].has_scenarios() \
                and not self.scenario_names:
            num_scenarios = self.data[0].num_scenarios()
            self.scenario_names = [f'e{i:02}' for i in range(num_scenarios)]

    def has_scenarios(self):
        """
        Check whether or not this time series has scenarios.

        :return: True when this time series has scenarios, otherwise False
        :rtype: bool
        """
        return (
            len(self.scenario_names) > 0 or
            (self.data and self.data[0].has_scenarios())
        )

    def total_values_per_item(self):
        """
        Get the total number of values per item:

         * A regular time series has one value per item
         * A scenario-based time series will return number of scenarios
         * A forecast with mean and ensembles will return 1 + number of ensembles

        :return: Total number of values per date-time in this time series
        :rtype: int
        """
        value_type = self.value_type()
        if value_type == ValueType.VALUE:
            return 1
        if value_type == ValueType.SCENARIOS:
            return len(self.scenario_names)
        if value_type == ValueType.MEAN_AND_SCENARIOS:
            return 1 + len(self.scenario_names)
        raise ValueError(
            f"Timeseries has unknown value type: {value_type}"
        )

    def value_type(self):
        """
        Return the value type of this time series.

        :return: The value type of this time series
        :rtype: ValueType
        """
        if self.has_data():
            return self.data[0].value_type()
        else:
            return ValueType.VALUE

    def begin(self):
        if self.data:
            return self.data[0].date
        else:
            raise ValueError("Timeseries has no values")

    def end(self):
        if self.data:
            return self.resolution >> self.data[-1].date
        else:
            raise ValueError("Timeseries has no values")

    def to_df(self, name=None, single_level_header=False):
        """
        Alias for :meth:`Timeseries.to_dataframe`. Convert this timeseries
        to a ``pandas.DataFrame``.

        :param name: Set a name for the value column, defaults to ``value``
        :type name: str, optional
        :param single_level_header: Set to True to use single-level header \
            in the DataFrame, defaults to False
        :type single_level_header: boolean, optional
        :return: A DataFrame
        :rtype: pandas.DataFrame
        :raises ImportError: When pandas is not installed on the system
        """
        return self.to_dataframe(
            name=name,
            single_level_header=single_level_header
        )

    def to_dataframe(self, name=None, single_level_header=False):
        """
        Convert this timeseries to a ``pandas.DataFrame``.

        :param name: Set a name for the value column, defaults to ``value``
        :type name: str, optional
        :param single_level_header: Set to True to use single-level header \
            in the DataFrame, defaults to False
        :type single_level_header: boolean, optional
        :return: A DataFrame
        :rtype: pandas.DataFrame
        :raises ImportError: When pandas is not installed on the system
        """
        return timeseries_to_dataframe(
            self,
            name=name,
            single_level_header=single_level_header
        )

    def validate(self):
        """
        Validate the time series.
        """
        # Check resolution
        if not self.resolution:
            raise ValueError("Timeseries has no resolution")
        r = self.resolution
        # Empty is OK
        if not self.data:
            return
        # Check all data points
        d = self.data[0].date
        for index, item in enumerate(self.data):
            # Validate values
            item.validate()
            # Check date vs resolution
            if not r.matches(item.date):
                raise ValueError(
                    f"Date-time not in Resolution: {item.date} at index {index}"
                )
            # Check if the current date
            if not d == item.date:
                raise ValueError(
                    f"Expected {d}, but got {item.date} at index {index}"
                )
            d = r >> d

    def print(self, file=sys.stdout):
        """
        Utility method to print a time series to any file handle (defaults
        to stdout).
        """
        print(f"Timeseries:", file=file)
        if self.curve:
            print(f"  Curve: {repr(self.curve)}", file=file)
        if self.instance:
            print(f"  Instance: {self.instance}", file=file)
        if self.scenario_names:
            print(f"  Scenarios: {self.scenario_names}", file=file)
        print(f"  Resolution: {self.resolution}", file=file)
        print(f"", file=file)
        for d in self.data:
            d.print(file=file)


# Time series list implementation with helpers


class TimeseriesList(list):
    """
    A list of Timeseries objects. All time series must have the same
    frequency.

    :param iterable: Any iterable of time series
    :type iterable: iterable
    """

    def __init__(self, iterable=(), frequency=None):
        # Initialize list
        super().__init__(iterable)
        # Get frequency
        self._frequency = frequency
        # Asserts
        _validate_timeseries_list(self)
        _check_and_get_frequency_list(self, self._frequency)

    @property
    def frequency(self):
        return self._frequency

    def to_df(self, single_level_header=False):
        """
        Alias for :meth:`Timeseries.to_dataframe`.

        Convert this TimeseriesList to a ``pandas.DataFrame`` where all time
        series are placed in its own column and are lined up with the date-time
        as index.

        :param single_level_header: Set to True to use single-level header \
            in the DataFrame, defaults to False
        :type single_level_header: boolean, optional
        :return: A DataFrame
        :rtype: pandas.DataFrame
        :raises ImportError: When pandas is not installed on the system
        """
        return self.to_dataframe(single_level_header=single_level_header)

    def to_dataframe(self, single_level_header=False):
        """
        Convert this TimeseriesList to a ``pandas.DataFrame`` where all time
        series are placed in its own column and are lined up with the date-time
        as index.

        :param single_level_header: Set to True to use single-level header \
            in the DataFrame, defaults to False
        :type single_level_header: boolean, optional
        :return: A DataFrame
        :rtype: pandas.DataFrame
        :raises ImportError: When pandas is not installed on the system
        """
        return timeseries_list_to_dataframe(
            self,
            single_level_header=single_level_header
        )

    def append(self, timeseries):
        # Asserts
        _validate_timeseries(timeseries)
        self._frequency = _check_and_get_frequency(timeseries, self._frequency)
        # Perform operation
        return super().append(timeseries)

    def extend(self, iterable):
        # Asserts
        _validate_timeseries_list(iterable)
        self._frequency = _check_and_get_frequency_list(iterable, self._frequency)
         # Perform operation
        return super().extend(iterable)

    def insert(self, index, timeseries):
        # Asserts
        _validate_timeseries(timeseries)
        self._frequency = _check_and_get_frequency(timeseries, self._frequency)
         # Perform operation
        return super().insert(index, timeseries)

    def __add__(self, rhs):
        # Asserts
        _validate_timeseries_list(rhs)
        self._frequency = _check_and_get_frequency_list(rhs, self._frequency)
        # Perform operation
        return TimeseriesList(list.__add__(self, rhs), frequency=self._frequency)

    def __iadd__(self, rhs):
        # Asserts
        _validate_timeseries_list(rhs)
        self._frequency = _check_and_get_frequency_list(rhs, self._frequency)
        # Perform operation
        return TimeseriesList(list.__iadd__(self, rhs), frequency=self._frequency)

    def __setitem__(self, key, timeseries):
        # Asserts
        _validate_timeseries(timeseries)
        self._frequency = _check_and_get_frequency(timeseries, self._frequency)
        # Perform operation
        return super().__setitem__(timeseries)

    def __mul__(self, rhs):
        raise NotImplementedError("TimeseriesList does not support multiply")

    def __rmul__(self, rhs):
        raise NotImplementedError("TimeseriesList does not support multiply")

    def __imul__(self, rhs):
        raise NotImplementedError("TimeseriesList does not support multiply")

    def copy(self):
        return TimeseriesList(self, frequency=self._frequency)

    def __getitem__(self, item):
        result = list.__getitem__(self, item)
        if isinstance(result, list):
            return TimeseriesList(result, frequency=self._frequency)
        else:
            return result


def _find_frequency(timeseries_list):
    if timeseries_list:
        return timeseries_list[0].resolution.frequency
    return None


def _check_and_get_frequency(timeseries, frequency=None):
    if frequency:
        assert timeseries.resolution.frequency == frequency, (
            f"Items in TimeseriesList must have frequency {frequency}, but "
            f"the Timeseries has {timeseries.resolution.frequency}."
        )


def _check_and_get_frequency_list(timeseries_list, frequency=None):
    if not frequency:
        frequency = _find_frequency(timeseries_list)
    if frequency:
        for index, timeseries in enumerate(timeseries_list):
            assert timeseries.resolution.frequency == frequency, (
                f"Element {index} does not match the frequency of "
                f"the other Timeseries objects."
            )


def _validate_timeseries(timeseries):
    assert isinstance(timeseries, Timeseries), (
        f"Element is not a Timeseries. Expects all "
        f"elements to be Timeseries objects."
    )


def _validate_timeseries_list(timeseries_list):
    for index, timeseries in enumerate(timeseries_list):
        assert isinstance(timeseries, Timeseries), (
            f"Element {index} is not a Timeseries. Expects all "
            f"elements to be Timeseries objects."
        )



from .frequency import Frequency
from .timezone import UTC, CET, EET, WET, TRT, GAS_DAY, DEFAULT_TZ, local_tz
from .resolution import Resolution
from .utils import now, today, to_timezone, get_date, get_datetime


__all__ = [
    # Resolution, Frequency and common time-zones
    "Resolution",
    "Frequency",
    "DEFAULT_TZ",
    "local_tz",
    "UTC",
    "CET",
    "EET",
    "WET",
    "TRT",
    "GAS_DAY",
    # Various helpers
    "now",
    "today",
    "to_timezone",
    "get_date",
    "get_datetime",
]





class Instance:
    """
    An instance identifies a forecast or any other time series that
    was issued at a specific time.

    It is typically used to specify a weather forecast that a specific
    forecast is based on.

    An instance is identified by the combination of ``(issued, tag)``.
    """

    def __init__(self, issued, tag='', scenarios=None,
                 created=None, modified=None):
        assert issued, "issued cannot be None"
        assert tag is not None, "tag cannot be None"
        #: The issue date of this instance
        self.issued = issued
        #: The tag for this instance
        self.tag = tag
        #: The number of scenarios available in this instance (default=0)
        self.scenarios = scenarios
        #: When this instance was created (if available)
        self.created = created
        #: When this instance was modified (if available)
        self.modified = modified

    def as_dataframe_column_header(self):
        """
        Create a string fitting for a column header in the a
        ``pandas.DataFrame``.

        :return: A DataFrame column header text for an instance
        :rtype: str
        """
        return f"{self.issued:%Y-%m-%d %H:%M} {self.tag}"

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        if self.scenarios:
            return (
                f"<Instance: issued=\"{self.issued}\", tag=\"{self.tag}\", "
                f"scenarios={self.scenarios}>"
            )
        else:
            return f"<Instance: issued=\"{self.issued}\", tag=\"{self.tag}\">"


import logging
import time

import requests
from requests.exceptions import ConnectionError, Timeout, HTTPError


log = logging.getLogger(__name__)


class Retry:
    """
    Retry rules with max retries, number of seconds delay between the
    retries with a backoff factor (multiplies the delay for each missed
    retry).

    If you don't want any backoff, just leave the backoff factor at 1.
    """

    def __init__(self, max_tries=3, delay=7.5, backoff=2):
        self.max_tries = max_tries
        self.delay = delay
        self.backoff = backoff

    def __call__(self, func):
        """
        Retries decorator.
        """
        def _wrapper():
            # Local variables
            tries, delay, backoff = self.max_tries, self.delay, self.backoff
            error = None
            # Retry
            while tries > 1:
                # Perform operation
                try:
                    r = func()
                    # Check status versus 5xx (server errors)
                    if 500 <= r.status_code < 600:
                        log.error(
                            "HTTP %s %s for URL: %s"
                            % (r.status_code, r.reason, r.url)
                        )
                        raise HTTPError(
                            "HTTP %s %s for URL: %s"
                            % (r.status_code, r.reason, r.url)
                        )
                    # Return response if all was OK
                    return r
                except (ConnectionError, Timeout, HTTPError) as e:
                    error = e
                # Print info
                attempt = self.max_tries - tries + 1
                log.info(
                    "Attempt %d/%d failed. Retrying in %.1fs."
                    % (attempt, self.max_tries, delay)
                )
                # Sleep and update retrier
                time.sleep(delay)
                tries = tries - 1
                delay = delay * backoff
            # Last try
            return func()
        return _wrapper





class APIError(Exception):
    """
    Base exception for all errors that may occur during an API call.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class HTTPError(APIError):
    """
    Base class for all HTTP errors. Mostly used to wrap request's HTTP errors.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class UnauthorizedError(APIError):
    """
    Authentication errors such as invalid API key or that the user is blocked.
    """

    def __init__(self, reason=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reason = reason or ""

    def __str__(self):
        return self.reason


class ValidationError(APIError):
    """
    Validation error of some kind (invalid parameter or combination
    of parameters).
    """

    def __init__(self, reason=None, parameter=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reason = reason
        self.parameter = parameter

    def __str__(self):
        if self.parameter:
            return f"Field=\"{self.parameter}\", message: {self.reason}"
        else:
            return self.reason


class ForbiddenError(APIError):
    """
    The resource is not accessible for the user.
    """

    def __init__(self, reason=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reason = reason or ""

    def __str__(self):
        return self.reason


class NotFoundError(APIError):
    """
    The resource was not found on the server.
    """

    def __init__(self, reason=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reason = reason or ""

    def __str__(self):
        return self.reason


class InternalServerError(APIError):
    """
    The server failed to process the request.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __str__(self):
        return (
            "Server failed to process the request. If the problem persists, "
            "please send an email to contact@energyquantified.com."
        )



from .base import BaseAPI

from ..exceptions import ValidationError
from ..metadata import CurveType
from ..parser.metadata import parse_instance_list
from ..parser.timeseries import parse_timeseries, parse_timeseries_list


# Tuple of supported values for Curve.curve_type in the instances API
CURVE_TYPES = (CurveType.INSTANCE,)


class InstancesAPI(BaseAPI):
    """
    Instance API operations. Access these operations via an
    instance of the :py:class:`energyquantified.EnergyQuantified` class:

       >>> eq = EnergyQuantified(api_key="aaaa-bbbb-cccc-dddd")
       >>> eq.timeseries.load(curve, begin, end)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._cache = {}

    def tags(self, curve):
        """
        Get all tags available for a curve.

        This operation works for curves with ``curve_type = INSTANCE`` only.

        :param curve: The curve or curve name
        :type curve: :py:class:`energyquantified.metadata.Curve`, str
        :return: A set available tags for provded curve
        :rtype: set
        """
        # Build URL
        safe_curve = self._urlencode_curve_name(curve, curve_types=CURVE_TYPES)
        url = f"/instances/{safe_curve}/tags/"
        # HTTP request (check cache first)
        if self._cache.get(url):
            return self._cache[url]
        response = self._get(url)
        self._cache[url] = set(response.json())
        return self._cache[url]

    def list(
            self,
            curve,
            tags=None,
            exlude_tags=None,
            limit=25,
            issued_at_latest=None,
            issued_at_earliest=None):
        """
        List instances for the curve. Does not load any time series data.

        Filter on the instance `tags` or `exclude_tags` (list of values
        allowed), or restrict the instance issue date by setting
        `issued_at_latest` and/or `issued_at_earliest`.

        Limit the maximum number of returned instances by setting `limit`
        anywhere between 1 and 25.

        This operation works for curves with ``curve_type = INSTANCE`` only.

        :param curve: The curve or curve name
        :type curve: :py:class:`energyquantified.metadata.Curve`, str
        :param tags: Filter by instance tags, defaults to None
        :type tags: list, str, optional
        :param exlude_tags: Exclude instance tags, defaults to None
        :type exlude_tags: list, str, optional
        :param limit: Number of instances returned, defaults to 25
        :type limit: int, optional
        :param issued_at_latest: Filter by issue date, defaults to None
        :type issued_at_latest: datetime, date, str, optional
        :param issued_at_earliest: Filter by issue date, defaults to None
        :type issued_at_earliest: datetime, date, str, optional
        :return: A list of :py:class:`energyquantified.metadata.Instance`
        :rtype: list
        """
        # Build URL
        safe_curve = self._urlencode_curve_name(curve, curve_types=CURVE_TYPES)
        url = f"/instances/{safe_curve}/list/"
        # Parameters
        params = {}
        self._add_str_list(params, "tags", tags)
        self._add_str_list(params, "exclude-tags", exlude_tags)
        self._add_int(params, "limit", limit, min=1, max=25, required=True)
        self._add_datetime(params, "issued-at-latest", issued_at_latest)
        self._add_datetime(params, "issued-at-earliest", issued_at_earliest)
        # HTTP request
        response = self._get(url, params=params)
        return parse_instance_list(response.json())

    def load(
            self,
            curve,
            tags=None,
            exlude_tags=None,
            limit=5,
            issued_at_latest=None,
            issued_at_earliest=None,
            frequency=None,
            aggregation=None,
            hour_filter=None,
            ensembles=False):
        """
        Load time series instances.

        Filter on the instance `tags` or `exclude_tags` (list of values
        allowed), or restrict the instance issue date by setting
        `issued_at_latest` and/or `issued_at_earliest`.

        When `ensembles` is `False`: Adjust the maximum number of returned
        time series instances by setting `limit` between 1 and 25.

        When `ensembles` is `True`: `limit` must be between 1 and 10.

        Supports aggregation by specifying a `frequency`. Optionally, you
        can specify the `aggregation` method (default is average) and the
        hourly filter (base, peak, etc.).

        This operation works for curves with ``curve_type = INSTANCE`` only.

        :param curve: The curve or curve name
        :type curve: :py:class:`energyquantified.metadata.Curve`, str
        :param tags: Filter by instance tags, defaults to None
        :type tags: list, str, optional
        :param exlude_tags: Exclude instance tags, defaults to None
        :type exlude_tags: list, str, optional
        :param limit: Maximum number of instances returned, defaults to 5
        :type limit: int, optional
        :param issued_at_latest: [description], defaults to None
        :type issued_at_latest: [type], optional
        :param issued_at_earliest: [description], defaults to None
        :type issued_at_earliest: [type], optional
        :param frequency: Set the preferred frequency for aggregations,\
            defaults to None
        :type frequency: Frequency, optional
        :param aggregation: The aggregation method (i.e. AVERAGE, MIN, MAX),\
            has no effect unless *frequency* is provided, defaults to AVERAGE
        :type aggregation: Aggregation, optional
        :param hour_filter: Filters on hours to include (i.e. BASE, PEAK),\
            has no effect unless *frequency* is provided, defaults to BASE
        :type hour_filter: Filter, optional
        :param ensembles: Whether or not to include ensembles where available,\
            defaults to False
        :type ensembles: bool, optional
        :return: A list :py:class:`energyquantified.data.Timeseries` instances
        :rtype: list
        """
        # Build URL
        safe_curve = self._urlencode_curve_name(curve, curve_types=CURVE_TYPES)
        if ensembles:
            url = f"/ensembles/{safe_curve}/"
        else:
            url = f"/instances/{safe_curve}/"
        # Parameters
        params = {}
        self._add_str_list(params, "tags", tags)
        self._add_str_list(params, "exclude-tags", exlude_tags)
        if ensembles:
            self._add_int(params, "limit", limit, min=1, max=25, required=True)
        else:
            self._add_int(params, "limit", limit, min=1, max=10, required=True)
        self._add_datetime(params, "issued-at-latest", issued_at_latest)
        self._add_datetime(params, "issued-at-earliest", issued_at_earliest)
        self._add_frequency(params, "frequency", frequency)
        if "frequency" in params:
            self._add_aggregation(params, "aggregation", aggregation)
            self._add_filter(params, "hour-filter", hour_filter)
        # HTTP request
        response = self._get(url, params=params)
        return parse_timeseries_list(response.json())

    def latest(
            self,
            curve,
            tags=None,
            issued_at_latest=None,
            frequency=None,
            aggregation=None,
            hour_filter=None,
            ensembles=False):
        """
        Get the latest time series instance with filtering on `tags` and
        `issued_at_latest`.

        Supports aggregation by specifying a `frequency`. Optionally, you
        can specify the `aggregation` method (default is average) and the
        hourly filter (base, peak, etc.).

        This operation works for curves with ``curve_type = INSTANCE`` only.

        :param curve: The curve or curve name
        :type curve: :py:class:`energyquantified.metadata.Curve`, str
        :param tags: Filter by instance tags, defaults to None
        :type tags: list, str, optional
        :param issued_at_latest: [description], defaults to None
        :type issued_at_latest: [type], optional
        :param frequency: Set the preferred frequency for aggregations,\
            defaults to None
        :type frequency: Frequency, optional
        :param aggregation: The aggregation method (i.e. AVERAGE, MIN, MAX),\
            has no effect unless *frequency* is provided, defaults to AVERAGE
        :type aggregation: Aggregation, optional
        :param hour_filter: Filters on hours to include (i.e. BASE, PEAK),\
            has no effect unless *frequency* is provided, defaults to BASE
        :type hour_filter: Filter, optional
        :param ensembles: Whether or not to include ensembles where available,\
            defaults to False
        :type ensembles: bool, optional
        :return: A time series instance
        :rtype: :py:class:`energyquantified.data.Timeseries`
        """

        # Build URL
        safe_curve = self._urlencode_curve_name(curve, curve_types=CURVE_TYPES)
        if ensembles:
            url = f"/ensembles/{safe_curve}/get/latest/"
        else:
            url = f"/instances/{safe_curve}/get/latest/"
        # Parameters
        params = {}
        self._add_str_list(params, "tags", tags)
        self._add_datetime(params, "issued-at-latest", issued_at_latest)
        self._add_frequency(params, "frequency", frequency)
        if "frequency" in params:
            self._add_aggregation(params, "aggregation", aggregation)
            self._add_filter(params, "hour-filter", hour_filter)
        # HTTP request
        response = self._get(url, params=params)
        return parse_timeseries(response.json())

    def get(
            self,
            curve,
            issued=None,
            tag="",
            frequency=None,
            aggregation=None,
            hour_filter=None,
            ensembles=False):
        """
        Get an instance specified by a `issued` (issue date) and `tag`. The
        default tag is blank and tags are case-insensitive.

        Supports aggregation by specifying a `frequency`. Optionally, you
        can specify the `aggregation` method (default is average) and the
        hourly filter (base, peak, etc.).

        This operation works for curves with ``curve_type = INSTANCE`` only.

        :param curve: The curve or curve name
        :type curve: :py:class:`energyquantified.metadata.Curve`, str
        :param issued: The issue date-time, defaults to None
        :type issued: datetime, date, str, optional
        :param tag: The instance tag, defaults to ""
        :type tag: str, optional
        :param frequency: Set the preferred frequency for aggregations,\
            defaults to None
        :type frequency: Frequency, optional
        :param aggregation: The aggregation method (i.e. AVERAGE, MIN, MAX),\
            has no effect unless *frequency* is provided, defaults to AVERAGE
        :type aggregation: Aggregation, optional
        :param hour_filter: Filters on hours to include (i.e. BASE, PEAK),\
            has no effect unless *frequency* is provided, defaults to BASE
        :type hour_filter: Filter, optional
        :param ensembles: Whether or not to include ensembles where available,\
            defaults to False
        :type ensembles: bool, optional
        :return: A time series instance
        :rtype: :py:class:`energyquantified.data.Timeseries`
        """
        # Build URL
        safe_curve = self._urlencode_curve_name(curve, curve_types=CURVE_TYPES)
        safe_issued = self._urlencode_datetime(issued, "issued")
        category = "ensembles" if ensembles else "instances"
        if tag:
            safe_tag = self._urlencode_string(tag, "tag", lambda t: t.lower())
            url = f"/{category}/{safe_curve}/get/{safe_issued}/{safe_tag}/"
        else:
            url = f"/{category}/{safe_curve}/get/{safe_issued}/"
        # Parameters
        params = {}
        self._add_frequency(params, "frequency", frequency)
        if "frequency" in params:
            self._add_aggregation(params, "aggregation", aggregation)
            self._add_filter(params, "hour-filter", hour_filter)
        # HTTP request
        response = self._get(url, params=params)
        return parse_timeseries(response.json())

    def relative(
            self,
            curve,
            begin=None,
            end=None,
            tag=None,
            days_ahead=1,
            issued="latest",
            time_of_day=None,
            after_time_of_day=None,
            before_time_of_day=None,
            frequency=None,
            aggregation=None,
            hour_filter=None):
        """
        Load one instance for each day based on some common criteria, stitch
        them together and return a continuous time series.

        By default, this method selects the day-ahead instances (forecasts),
        but you can set ``days_ahead`` to 0 or higher. 0 means intraday,
        1 means the day-ahead (default), 2 means the day after day-ahead,
        and so on.

        You may control the time of the day the instance is issued by setting
        exactly one of the follow parameters: ``time_of_day``,
        ``after_time_of_day`` or ``before_time_of_day``. These should be set
        to a time (HH:MM:SS). You can use the :py:class:`datetime.time`.

        This operation works for curves with ``curve_type = INSTANCE`` only.

        :param curve: The curve or curve name
        :type curve: :py:class:`energyquantified.metadata.Curve`, str
        :param begin: The begin date-time
        :type begin: date, datetime, str, required
        :param end: The end date-time
        :type end: date, datetime, str, required
        :param days_ahead: The number of leading days (0 or higher),\
            defaults to 1
        :type days_ahead: int, optional
        :param issued: Whether to select the earliest or latest matching\
            instance per day, allowed values "earliest" | "latest",\
            defaults to "earliest"
        :type issued: str, optional
        :param time_of_day: The exact time of the instance is issued,\
            defaults to None
        :type time_of_day: time, str, optional
        :param after_time_of_day: The instance must be issued at this time of\
            day or after, defaults to None
        :type after_time_of_day: time, str, optional
        :param before_time_of_day:  The instance must be issued before this\
            time of day, defaults to None
        :type before_time_of_day: time, str, optional
        :param frequency: Set the preferred frequency for aggregations,\
            defaults to None
        :type frequency: Frequency, optional
        :param aggregation: The aggregation method (i.e. AVERAGE, MIN, MAX),\
            has no effect unless *frequency* is provided, defaults to AVERAGE
        :type aggregation: Aggregation, optional
        :param hour_filter: Filters on hours to include (i.e. BASE, PEAK),\
            has no effect unless *frequency* is provided, defaults to BASE
        :type hour_filter: Filter, optional
        :return: A time series
        :rtype: :py:class:`energyquantified.data.Timeseries`
        """
        # Build URL
        safe_curve = self._urlencode_curve_name(curve, curve_types=CURVE_TYPES)
        url = f"/instances/{safe_curve}/get/relative/"
        # Parameters
        params = {}
        self._add_datetime(params, "begin", begin, required=True)
        self._add_datetime(params, "end", end, required=True)
        self._add_str(params, "tag", tag, required=True)
        self._add_int(params, "days-ahead", days_ahead, min=0, max=10000, required=True)
        self._add_str(params, "issued", issued, allowed_values=['earliest', 'latest'], required=True)
        self._add_time(params, "time-of-day", time_of_day)
        self._add_time(params, "after-time-of-day", after_time_of_day)
        self._add_time(params, "before-time-of-day", before_time_of_day)
        self._add_frequency(params, "frequency", frequency)
        if "frequency" in params:
            self._add_aggregation(params, "aggregation", aggregation)
            self._add_filter(params, "hour-filter", hour_filter)
        # Additional validation checks
        if sum(1 if t is not None else 0 for t in (time_of_day, after_time_of_day, before_time_of_day)) > 1:
            raise ValidationError(reason=(
                "At most one of the following fields must be set: "
                "time_of_day, after_time_of_day, before_time_of_day"
            ))
        # HTTP request
        response = self._get(url, params=params)
        return parse_timeseries(response.json())



import os
from setuptools import setup, find_packages


here = os.path.abspath(os.path.dirname(__file__))


# Requirements
install_requires = [
    "requests==2.25.1",
    "pytz",
    "tzlocal",
    "python-dateutil==2.8.1",
]
# tests_require = [
#     "pytest",
#     "requests-mock",
# ]

# Get package info
about = {}
package_info_filename = os.path.join(here, "energyquantified", "__version__.py")
with open(package_info_filename, "r", encoding="utf-8") as f:
    exec(f.read(), about)

# Get the README.md file (long description)
with open(os.path.join(here, "README.md"), "r", encoding="utf-8") as f:
    readme = f.read()

# Classifiers
classifiers = [
    "Development Status :: 4 - Beta",
    "Environment :: Console",
    "Intended Audience :: Developers",
    "Intended Audience :: Science/Research",
    "Intended Audience :: Financial and Insurance Industry",
    "License :: OSI Approved :: Apache Software License",
    "Natural Language :: English",
    "Operating System :: Microsoft :: Windows",
    "Operating System :: POSIX :: Linux",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Topic :: Office/Business :: Financial",
    "Topic :: Software Development :: Libraries :: Python Modules",
]


setup(
    name=about["__title__"],
    version=about["__version__"],
    description=about["__description__"],
    long_description=readme,
    long_description_content_type="text/markdown",
    author=about["__author__"],
    url=about["__url__"],
    packages=find_packages(include=["energyquantified", "energyquantified.*"]),
    package_data={
        '': ["LICENSE"],
    },
    python_requires=">=3.7",
    install_requires=install_requires,
    # tests_require=tests_require,
    project_urls={
        "Documentation": "https://energyquantified-python.readthedocs.io",
        "Source": "https://github.com/energyquantified/eq-python-client",
    },
    classifiers=classifiers
)



import pytz
import tzlocal

from ._timezone_gas import _build_europe_gas_day_tzinfo


# Look up the most used time zones in the European power markets
# from the pytz library

UTC = pytz.UTC
CET = pytz.timezone("CET")
WET = pytz.timezone("WET")
EET = pytz.timezone("EET")
TRT = pytz.timezone("Europe/Istanbul")
GAS_DAY = _build_europe_gas_day_tzinfo()


def local_tz():
    """
    Get the local time-zone.

    :return: The time-zone for this system.
    :rtype: TzInfo
    """
    return tzlocal.get_localzone()


# Default time zone

DEFAULT_TZ = CET
LOCAL_TZ = local_tz()



import enum
from datetime import datetime, date

from ..time import Frequency


_aggregation_lookup = {}

class Aggregation(enum.Enum):
    """
    Supported aggregations in the API. Includes simple implementations of
    the different aggregation types for learning purposes.
    """

    #: Calculate the mean value
    AVERAGE = (lambda items: sum(items) / len(items),)
    #: Calculate the sum of all values
    SUM = (sum,)
    #: Find the minimum value
    MIN = (min,)
    #: Find the maximum value
    MAX = (max,)

    def __init__(self, func):
        self._func = func
        _aggregation_lookup[self.tag.lower()] = self

    @property
    def tag(self):
        """
        Get the tag for this aggregation type.

        :return: The aggregation tag (name)
        :rtype: str
        """
        return self.name

    def aggregate(self, iterable):
        """
        Perform an aggregation on any iterable of numbers (such as lists,
        tuples, generators etc.).

        :param iterable: Any iterable of numbers
        :type iterable: iterable
        :return: An aggregate
        :rtype: float
        """
        return self._func(iterable)

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    @staticmethod
    def is_valid_tag(tag):
        """
        Check whether an aggregation tag exists or not.

        :param tag: An aggregation tag
        :type tag: str
        :return: True if it exists, otherwise False
        :rtype: bool
        """
        return tag.lower() in _aggregation_lookup

    @staticmethod
    def by_tag(tag):
        """
        Look up aggregation by tag.

        :param tag: An aggregation tag
        :type tag: str
        :return: The aggregation for the given tag
        :rtype: Aggregation
        """
        return _aggregation_lookup[tag.lower()]


_filter_lookup = {}

class Filter(enum.Enum):
    """
    Supported filters in the API.

    Includes simple implementations of the filters for learning
    purposes.

    Note: The API automatically separates futures peak and offpeak by
    looking at the selected frequency for aggregations:

     * For weekly, monthly, quartly and yearly frequency, the futures
       peak and offpeak are used.
     * For daily or higher frequencies, the standard peak and offpeak
       are used.

    """

    #: All hours
    BASE = (
        lambda dt: True,
        lambda dt: True
    )
    #: Peak hours
    PEAK = (
        lambda dt: 8 <= dt.hour <= 19,
        lambda dt: dt.isoweekday() <= 5 and 8 <= dt.hour <= 19
    )
    #: Offpeak hours
    OFFPEAK = (
        lambda dt: not PEAK.is_in_filter(dt),
        lambda dt: not PEAK.is_in_future_filter(dt),
    )
    #: Monday–Friday
    WORKDAYS = (
        lambda dt: dt.isoweekday() <= 5,
        lambda dt: dt.isoweekday() <= 5
    )
    #: Saturday and Sunday
    WEEKEND = (
        lambda dt: dt.isoweekday() > 5,
        lambda dt: dt.isoweekday() > 5
    )

    def __init__(self, filter_func, future_filter_func):
        self._filter_func = filter_func
        self._future_filter_func = future_filter_func
        _filter_lookup[self.tag.lower()] = self

    @property
    def tag(self):
        """
        The filter tag (name)
        """
        return self.name

    def get_filter_function(self, frequency):
        """
        Given a frequency, return the appropriate filter function:

         * For weekly, monthly, quartly and yearly frequency, the futures
           peak and offpeak function is used.
         * For daily or higher frequencies, the standard peak and offpeak
           function is used.

        :param frequency: The resulting frequency of an aggregation
        :type frequency: Frequency
        :return: A filter function
        :rtype: function
        """
        assert isinstance(frequency, Frequency), "Not a Frequency provided"
        if frequency >= Frequency.P1D:
            # For daily, hourly or higher frequencies
            return self._filter_func
        else:
            # For futures resolutions (weekly, monthly, quarterly, yearly)
            return self._future_filter_func

    def is_in_filter(self, datetime_obj):
        """
        Check whether or not a datetime object is in a filter.

        :param datetime_obj: A date-time
        :type datetime_obj: datetime
        :return: True if the date-time object falls into filter, otherwise False
        :rtype: bool
        """
        assert isinstance(datetime_obj, datetime), "Not a datetime object"
        return self._filter_func(datetime_obj)

    def is_in_future_filter(self, datetime_obj):
        """
        Check whether or not a datetime object is in a filter for
        futures-contracts (weekly, monthly, quarterly, yearly).

        :param datetime_obj: A date-time
        :type datetime_obj: datetime
        :return: True if the date-time object falls into filter, otherwise False
        :rtype: bool
        """
        assert isinstance(datetime_obj, (datetime, date)), (
            "Not a datetime or date object"
        )
        return self._future_filter_func(datetime_obj)

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    @staticmethod
    def is_valid_tag(tag):
        """
        Check whether a filter tag exists or not.

        :param tag: A filter tag
        :type tag: str
        :return: True if it exists, otherwise False
        :rtype: bool
        """
        return tag.lower() in _filter_lookup

    @staticmethod
    def by_tag(tag):
        """
        Look up a filter by tag.

        :param tag: A filter tag
        :type tag: str
        :return: The filter for the given tag
        :rtype: Filter
        """
        return _filter_lookup[tag.lower()]



from dataclasses import dataclass
from datetime import date
import enum

from ..time import Frequency


_periods_lookup = {}

class ContractPeriod(enum.Enum):
    """
    Enumerator of contract periods available in for OHLC contracts
    in the API. The contract periods are:

     * ``YEAR`` – Yearly contract period
     * ``MDEC`` – December delivery (for ETS EUA contracts)
     * ``SEASON`` – Summer or Winter (typically for gas contracts)
     * ``QUARTER`` – Quarterly contract period
     * ``MONTH`` – Monthly contract period
     * ``WEEK`` – Weekly contract period
     * ``WEEKEND`` – Weekend contract period
     * ``DAY`` – Daily contract period
    """

    # Contract periods for OHLC data
    YEAR = ("year", Frequency.P1Y)
    MDEC = ("mdec", Frequency.P1Y)
    SEASON = ("season", Frequency.SEASON)
    QUARTER = ("quarter", Frequency.P3M)
    MONTH = ("month", Frequency.P1M)
    WEEK = ("week", Frequency.P1W)
    WEEKEND = ("weekend", Frequency.P1D)
    DAY = ("day", Frequency.P1D)

    def __init__(self, tag, frequency):
        self.tag = tag
        self.frequency = frequency
        _periods_lookup[self.tag.lower()] = self

    @staticmethod
    def by_tag(tag):
        """
        Look up contract periods by tag.

        :param tag: The tag to look up
        :type tag: str
        :return: A ContractPeriod for this tag
        :rtype: ContractPeriod
        """
        return _periods_lookup[tag.lower()]

    @staticmethod
    def is_valid_tag(tag):
        """
        Check whether a contract period tag exists or not.

        :param tag: The tag to look up
        :type tag: str
        :return: True if the contract period tag exists, otherwise False
        :rtype: bool
        """
        return tag.lower() in _periods_lookup

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


_ohlcfield_lookup = {}

class OHLCField(enum.Enum):
    """
    A field in an OHLC object. Used to specify which field to load or
    to perform an operation on.

     * ``OPEN`` – Opening trade for the trading day
     * ``HIGH`` – Highest trade for the trading day
     * ``LOW`` – Lowest trade for the trading day
     * ``CLOSE`` – Closing trade for the trading day
     * ``SETTLEMENT`` – Settlement price for the trading day
     * ``VOLUME`` – Accumulated traded volume for the trading day
     * ``OPEN_INTEREST`` – Total volume of contracts that have
       been entered into and not yet liquidated by an offsetting
       transaction or fulfilled by delivery
    """

    OPEN = ("open",)
    HIGH = ("high",)
    LOW = ("low",)
    CLOSE = ("close",)
    SETTLEMENT = ("settlement",)
    VOLUME = ("volume",)
    OPEN_INTEREST = ("open_interest",)

    def __init__(self, tag):
        self.tag = tag
        _ohlcfield_lookup[tag.lower()] = self

    def get_value(self, ohlc):
        """
        Get the value from an OHLC object for this field.

        :param ohlc: An OHLC object
        :type ohlc: OHLC
        :return: The value for that given OHLC field
        :rtype: float
        """
        return getattr(ohlc, self.tag, default=None)

    @staticmethod
    def by_tag(tag):
        """
        Look up OHLC field by tag.

        :param tag: The tag to look up
        :type tag: str
        :return: An OHLCField for this tag
        :rtype: OHLCField
        """
        return _ohlcfield_lookup[tag.lower()]

    @staticmethod
    def is_valid_tag(tag):
        """
        Check whether a OHLC field tag exists or not.

        :param tag: The tag to look up
        :type tag: str
        :return: True if the OHLC field tag exists, otherwise False
        :rtype: bool
        """
        return tag.lower() in _ohlcfield_lookup

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


@dataclass(frozen=True)
class ContinuousContract:
    """
    Metadata class for keeping track of which contract was queried. Continuous
    contracts are the rolling `N`-front contract – relative to its trading day.

    For instance, a contract with ``period=MONTH``, ``front=1`` is what you
    would typically call a front month contract. ``period=MONTH``, ``front=2``
    is the second closest front contract, and so forth.

    See also :py:class:`SpecificContract`.

    .. py:attribute:: field

        The field to have fetched from the OHLC objects (such as close,
        settlement, open, high, low, volume)

    .. py:attribute:: period

        The contract period (i.e. week, month, quarter)

    .. py:attribute:: front

        The front number (1 is the closest contract, 2 is the second closest,
        and so on)
    """
    period: ContractPeriod
    front: int
    field: OHLCField = None

    def is_continuous(self):
        """
        Check whether this contract instance is continuous.

        :return: True if continuous, otherwise False
        :rtype: bool
        """
        return True

    def is_specific(self):
        """
        Check whether this contract instance is specific.

        :return: True if specific, otherwise False
        :rtype: bool
        """
        return False

    def as_dataframe_column_header(self):
        """
        Create a string fitting for a column header in the a
        ``pandas.DataFrame``.

        :return: A DataFrame column header text for this contract
        :rtype: str
        """
        return f"{self.period.tag} front-{self.front} {self.field.tag}"

    def __str__(self):
        return (
            f"<ContinuousContract: "
            f"period={self.period.name}, "
            f"front={self.front}, "
            f"field={self.field.name if self.field else ''}>"
        )

    def __repr__(self):
        return self.__str__()


@dataclass(frozen=True)
class SpecificContract:
    """
    Metadata class for keeping track of which contract was queried. Specific
    contracts are contracts which goes to delivery on a specific date.

    For instance, a contract with ``period=MONTH``, ``delivery=2020-12-01``
    is the contract for delivery in December 2020.

    See also :py:class:`ContinuousContract`.

    .. py:attribute:: field

        The field to have fetched from the OHLC objects (such as close,
        settlement, open, high, low, volume)

    .. py:attribute:: period

        The contract period (i.e. week, month, quarter)

    .. py:attribute:: delivery

        The delivery date of the contract
    """
    period: ContractPeriod
    delivery: date
    field: OHLCField = None

    def is_continuous(self):
        """
        Check whether this contract instance is continuous.

        :return: True if continuous, otherwise False
        :rtype: bool
        """
        return False

    def is_specific(self):
        """
        Check whether this contract instance is specific.

        :return: True if specific, otherwise False
        :rtype: bool
        """
        return True

    def as_dataframe_column_header(self):
        """
        Create a string fitting for a column header in the a
        ``pandas.DataFrame``.

        :return: A DataFrame column header text for this contract
        :rtype: str
        """
        return f"{self.period.tag} {self.delivery.isoformat()} {self.field.tag}"

    def __str__(self):
        return (
            f"<SpecificContract: "
            f"period={self.period.name}, "
            f"delivery={self.delivery.isoformat()}, "
            f"field={self.field.name if self.field else ''}>"
        )

    def __repr__(self):
        return self.__str__()



import pytz
from dateutil import parser

from ..exceptions import ParseException
from ..metadata import (
    Curve, Instance, Area, DataType, CurveType, Place, PlaceType,
    ContractPeriod, ContinuousContract, SpecificContract, OHLCField
)
from ..time import Frequency, Resolution, UTC, to_timezone
from ..time.timezone import LOCAL_TZ


def parse_resolution(json):
    """
    Parse a JSON response from the server into a Resolution object.
    """
    frequency = Frequency.by_tag(json.get("frequency"))
    timezone = pytz.timezone(json.get("timezone"))
    return Resolution(frequency, timezone)


def parse_curve(json):
    """
    Parse a JSON response from the server into a Curve object.
    """
    name = json.get("name")
    curve_type = CurveType.by_tag(json.get("curve_type"))
    area = Area.by_tag(json.get("area"))
    categories = json.get("categories")
    resolution = json.get("resolution")
    frequency = Frequency.by_tag(resolution.get("frequency"))
    timezone = pytz.timezone(resolution.get("timezone"))
    unit = json.get("unit")
    denominator = json.get("denominator")
    data_type = DataType.by_tag(json.get("data_type"))
    source = json.get("source")
    commodity = json.get("commodity")

    area_sink = json.get("area_sink")
    if area_sink:
        area_sink = Area.by_tag(area_sink)

    place = json.get("place")
    if place:
        place = parse_place(place)

    instance_issued_timezone = json.get("instance_issued_timezone")
    if instance_issued_timezone:
        instance_issued_timezone = pytz.timezone(instance_issued_timezone)

    return Curve(
        name,
        curve_type=curve_type,
        instance_issued_timezone=instance_issued_timezone,
        area=area,
        area_sink=area_sink,
        place=place,
        frequency=frequency,
        timezone=timezone,
        categories=categories,
        unit=unit,
        denominator=denominator,
        data_type=data_type,
        source=source,
        commodity=commodity
    )


def parse_instance_list(json, curve=None):
    """
    Parse a JSON list-resonse from the server into a list of Instance objects.
    """
    if not isinstance(json, list):
        raise ParseException(
            f"Expected list of Instance JSON objects, found: {type(json)}"
        )
    return [parse_instance(json_obj) for json_obj in json]


def parse_instance(json, curve=None):
    """
    Parse a JSON resonse from the server into an Instance object.
    """
    # Find timezone
    if curve and curve.instance_issued_timezone:
        timezone = curve.instance_issued_timezone
    else:
        timezone = UTC
    # Parse the issue date
    issued = to_timezone(
        parser.isoparse(json.get("issued")),
        tz=timezone
    )
    tag = json.get("tag") or ""
    scenarios = json.get("scenarios") or None
    # Created and modified
    created = json.get("created") or None
    if created:
        created = to_timezone(
            parser.isoparse(created),
            tz=LOCAL_TZ
        )
    modified = json.get("modified") or None
    if modified:
        modified = to_timezone(
            parser.isoparse(modified),
            tz=LOCAL_TZ
        )
    return Instance(
        issued,
        tag,
        scenarios=scenarios,
        created=created,
        modified=modified
    )


def parse_place(json):
    """
    Parse a JSON resonse from the server into a Place object.
    """
    kind = PlaceType.by_tag(json.get("type"))
    key = json.get("key")
    name = json.get("name")
    unit = json.get("unit")
    fuels = json.get("fuels") or []
    location = json.get("location") or None

    area = json.get("area")
    if area:
        area = Area.by_tag(area)

    areas = json.get("areas")
    if areas:
        areas = [Area.by_tag(a) for a in areas]
    else:
        areas = []

    children = json.get("children")
    if children:
        children = [parse_place(p) for p in children]

    curves = json.get("curves")
    if curves:
        curves = [parse_curve(c) for c in curves]

    return Place(
        kind,
        key,
        name,
        unit=unit,
        fuels=fuels,
        area=area,
        areas=areas,
        location=location,
        children=children,
        curves=curves,
    )


def parse_contract(json):
    contract_type = json.get("type")

    field = json.get("field")
    if field:
        field = OHLCField.by_tag(field)
    period = ContractPeriod.by_tag(json.get("period"))

    if contract_type == "CONTINUOUS":
        front = json.get("front")
        return ContinuousContract(
            field=field,
            period=period,
            front=front
        )

    if contract_type == "SPECIFIC":
        delivery = parser.isoparse(json.get("delivery")).date()
        return SpecificContract(
            field=field,
            period=period,
            delivery=delivery
        )

    raise ParseException(
        f"Unknown contract.type in JSON: {contract_type}"
    )



from .base import BaseAPI

from ..metadata import CurveType
from ..parser.periodseries import parse_periodseries


# Tuple of supported values for Curve.curve_type in the periods API
CURVE_TYPES = (CurveType.PERIOD,)


class PeriodsAPI(BaseAPI):
    """
    Period-based series API operations. Access these operations via an
    instance of the :py:class:`energyquantified.EnergyQuantified` class:

       >>> eq = EnergyQuantified(api_key="aaaa-bbbb-cccc-dddd")
       >>> eq.periods.load(curve, begin, end)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def load(
            self,
            curve,
            begin=None,
            end=None):
        """
        Load period-based series data for a curve.

        This operation works for curves with ``curve_type = PERIOD`` only.

        :param curve: The curve or curve name
        :type curve:  :py:class:`energyquantified.metadata.Curve`, str
        :param begin: The begin date-time
        :type begin: date, datetime, str, required
        :param end: The end date-time
        :type end: date, datetime, str, required
        :return: A period-based series
        :rtype: :py:class:`energyquantified.data.Periodseries`
        """
        # Build URL
        safe_curve = self._urlencode_curve_name(curve, curve_types=CURVE_TYPES)
        url = f"/periods/{safe_curve}/"
        # Parameters
        params = {}
        self._add_datetime(params, "begin", begin, required=True)
        self._add_datetime(params, "end", end, required=True)
        # HTTP request
        response = self._get(url, params=params)
        return parse_periodseries(response.json())



# Basic package info

__title__ = "energyquantified"
__description__ = "Energy Quantified Time series API client."
__url__ = "https://github.com/energyquantified/eq-python-client"
__version__ = "0.7.1"
__author__ = "Energy Quantified AS"
__license__ = "Apache License 2.0"
__copyright__ = "Copyright \u00A9 Energy Quantified AS"



import enum

from ..time import Resolution, UTC


class Curve:
    """
    The curve identifies any type of time series data and OHLC data.

    The ``curve.name`` is used in the API when loading data for a curve.
    """

    def __init__(self, name, curve_type=None, instance_issued_timezone=None,
                 area=None, area_sink=None, place=None,
                 resolution=None, frequency=None, timezone=None,
                 categories=None, unit=None, denominator=None, data_type=None,
                 source=None, commodity=None):
        #: The curve name is the identifier.
        self.name = name
        #: Curve type (the type of data this curve refers to).
        self.curve_type = curve_type
        if self.curve_type.has_instances:
            #: For instance-based curves: The time-zone of the issue date
            #: in the instance, see :py:attr:`Instance.issued`.
            self.instance_issued_timezone = instance_issued_timezone or UTC
        else:
            self.instance_issued_timezone = None
        # The areas and place (if any)
        #: The area
        self.area = area
        #: The importing area for exchange curves
        self.area_sink = area_sink
        if area_sink:
            #: The exporting area for exchange curves
            self.area_source = area
        self.place = place
        # Resolution
        if resolution:
            #: The frequency of data in this curve
            self.frequency = resolution.frequency
            #: The time-zone of date-times in this curve
            self.timezone = resolution.timezone
        else:
            self.frequency = frequency
            self.timezone = timezone
        # Other metadata
        #: List of categories for this curve.
        self.categories = categories
        #: The unit (MW, EUR, etc.). See also :py:attr:`Curve.denominator`.
        self.unit = unit
        #: The denominator (for EUR/MWh: unit=EUR and denominator=MWh). See
        #: also :py:attr:`Curve.unit`.
        self.denominator = denominator
        #: The data type, :py:class:`DataType`.
        self.data_type = data_type
        #: The source of the data.
        self.source = source
        #: The curve commodity (Power, Gas, etc.)
        self.commodity = commodity

    @property
    def resolution(self):
        """
        The resolution (combination of frequency and timezone) for this curve.
        """
        return Resolution(self.frequency, self.timezone)

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"<Curve: \"{self.name}\", curve_type={self.curve_type}>"


_datatype_lookup = {}

class DataType(enum.Enum):
    """
    Data types describe the type of data (i.e. actuals, forecast). This is
    the attribute that is always set as the last word in the curve name.
    """

    #: Third-party actuals collected by Energy Quantified, but not modified.
    ACTUAL = ("ACTUAL", "Actual")
    #: Scenario data generated by Energy Quantified, which is based on climate
    #: data sets (synthetic weather years).
    CLIMATE = ("CLIMATE", "Climate")
    #: Scenario data generated by Energy Quantified. If you are looking for
    #: weather-based scenarios, look at ``DataType.CLIMATE``.
    SCENARIO = ("SCENARIO", "Scenario")
    #: A combination of third-party actuals and numbers generated by Energy
    #: Quantified, where we have filled missing with our best calculations.
    SYNTHETIC = ("SYNTHETIC", "Synthetic")
    #: The forecast models run backwards.
    BACKCAST = ("BACKCAST", "Backcast")
    #: The seasonal normals using 40 weather years.
    NORMAL = ("NORMAL", "Normal")
    #: Some model value (such as a factor).
    VALUE = ("VALUE", "Value")
    #: Forecasts generated by Energy Quantified unless another source is
    #: explicitly stated in the curve name.
    FORECAST = ("FORECAST", "Forecast")
    #: Currency rates.
    FOREX = ("FOREX", "Forex")
    #: Closing data from the market.
    OHLC = ("OHLC", "OHLC")
    #: Capacity data generated from REMIT outage messages.
    REMIT = ("REMIT", "REMIT")
    #: Total installed capacity.
    CAPACITY = ("CAPACITY", "Capacity")

    def __init__(self, tag=None, label=None):
        self.tag = tag
        self.label = label
        _datatype_lookup[tag.lower()] = self

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return self.name

    @staticmethod
    def is_valid_tag(tag):
        """
        Check whether a data type tag exists or not.

        :param tag: A data type tag
        :type tag: str
        :return: True if it exists, otherwise False
        :rtype: bool
        """
        return tag.lower() in _datatype_lookup

    @staticmethod
    def by_tag(tag):
        """
        Look up data type by tag.

        :param tag: A data type tag
        :type tag: str
        :return: The data type for the given tag
        :rtype: DataType
        """
        return _datatype_lookup[tag.lower()]


_curvetype_lookup = {}

class CurveType(enum.Enum):
    """
    Curve type is not a part of the curve name.

    Curve type describes the storage format of the underlying data and which
    operations must be used to fetch data for these curves.

     * Load time series and scenario-based time series using the
       ``EnergyQuantified.timeseries.*`` operations.
     * To load instances (i.e. forecasts), use the
       ``EnergyQuantified.timeseries.*`` operations.
     * Periods and period-instances can be loaded by using each of
       their respective operations located under
       ``EnergyQuantified.periods.*`` and
       ``EnergyQuantified.instance_periods.*``.
     * OHLC means "open, high, low and close" data. To load data from
       these curves, use the OHLC operations.
    """

    #: Plain, fixed-interval time series data
    TIMESERIES = ("TIMESERIES", False)
    #: Plain, fixed-interval scenarios of time series data
    SCENARIO_TIMESERIES = ("SCENARIO_TIMESERIES", False)
    #: Instances (forecasts)
    INSTANCE = ("INSTANCE", True)
    #: Period-based data
    PERIOD = ("PERIOD", False)
    #: Instances of period-based data
    INSTANCE_PERIOD = ("INSTANCE_PERIOD", True)
    #: Closing prices for market data
    OHLC = ("OHLC", False)

    def __init__(self, tag=None, has_instances=False):
        self.tag = tag
        self.has_instances = has_instances
        _curvetype_lookup[tag.lower()] = self

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    @staticmethod
    def is_valid_tag(tag):
        """
        Check whether a curve type tag exists or not.

        :param tag: A curve type tag
        :type tag: str
        :return: True if it exists, otherwise False
        :rtype: bool
        """
        return tag.lower() in _curvetype_lookup

    @staticmethod
    def by_tag(tag):
        """
        Look up curve type by tag.

        :param tag: A curve type tag
        :type tag: str
        :return: The curve type for the given tag
        :rtype: CurveType
        """
        return _curvetype_lookup[tag.lower()]



from .http import Session

from .api import (
    InstancesAPI,
    MetadataAPI,
    RealtoMetadataAPI,
    TimeseriesAPI,
    PeriodsAPI,
    PeriodInstancesAPI,
    OhlcAPI,
    SrmcAPI,
)
from .exceptions import UnauthorizedError, InitializationError


# Defaults

BASE_PATH = "https://app.energyquantified.com/api"


class EnergyQuantified:
    """
    The Time Series API client.

    Wraps the Energy Quantified Time Series API. Handles validation, network
    errors, rate limiting, and parsing of the API responses.

    Exactly one of **api_key** and **api_key_file** must be set at
    the same time. **api_key_file** shall be the file path to a file
    that contains just the API key (blank lines a stripped when read).

    :param api_key: The API key for your user account
    :type api_key: string, optional
    :param api_key_file: A file path to a file that contains the API key
    :type api_key_file: string, file, optional
    :param ssl_verify: Whether or not to verify the server certificate,\
                       defaults to True
    :type ssl_verify: bool, optional
    :param timeout: Maximum timeout per HTTP request, defaults to 15.0
    :type timeout: float, optional
    :param http_delay: The minimum number of seconds between the start of\
                       each HTTP request, defaults to 0.125 seconds (8 req/s)
    :type http_delay: float, optional

    **Basic usage:**

       >>> from energyquantified import EnergyQuantified
       >>> eq = EnergyQuantified(api_key="aaaa-bbbb-cccc-dddd")
       >>> eq.metadata.curves(q="de wind power actual")

    """

    def __init__(
            self,
            api_key=None,
            api_key_file=None,
            ssl_verify=True,
            timeout=15.0,
            http_delay=0.125,
            api_url=BASE_PATH
        ):
        # Simple validations
        assert api_key or api_key_file, "api_key is missing"
        if api_key:
            assert not api_key_file, (
                "Exactly one of 'api_key' and 'api_key_file' "
                "must be set at the same time, but both were given"
            )
        if api_key_file:
            assert not api_key, (
                "Exactly one of 'api_key' and 'api_key_file' "
                "must be set at the same time, but both were given"
            )
        assert timeout >= 0, "timeout must be larger than 0s"
        assert http_delay >= 0.0667, (
            "http_delay must be 0.0667s or slower (15 req/s)"
        )
        assert api_url, "api_url is missing"
        # Attributes
        self._api_key = _find_api_key(api_key, api_key_file)
        self._api_url = api_url
        # HTTP client
        self._session = Session(
            timeout=timeout,
            verify=ssl_verify,
            base_url=self._api_url,
            headers={
                "X-API-Key": self._api_key,
            },
            delay=http_delay
        )
        # --- Public members ---
        #: See :py:class:`energyquantified.api.MetadataAPI`. For metadata
        #: queries (such as curve search and place lookups).
        self.metadata = MetadataAPI(self)
        #: See :py:class:`energyquantified.api.TimeseriesAPI`. For loading
        #: time series data.
        self.timeseries = TimeseriesAPI(self)
        #: See :py:class:`energyquantified.api.InstancesAPI`. For loading
        #: time series instances.
        self.instances = InstancesAPI(self)
        #: See :py:class:`energyquantified.api.PeriodsAPI`. For
        #: loading period-based series.
        self.periods = PeriodsAPI(self)
        #: See :py:class:`energyquantified.api.PeriodInstancesAPI`. For
        #: loading instances of period-based series.
        self.period_instances = PeriodInstancesAPI(self)
        #: See :py:class:`energyquantified.api.OhlcAPI`. For loading
        #: OHLC data in various ways
        self.ohlc = OhlcAPI(self)
        #: See :py:class:`energyquantified.api.SrmcAPI`. For loading and
        #: calculating short-run marginal costs (SRMC) from OHLC data.
        self.srmc = SrmcAPI(self)

    def is_api_key_valid(self):
        """
        Check if the supplied API key is valid/user account can sign in.

        Does so by trying to load the categories list. If it succeeds, then
        the user has supplied a valid API key.

        :return: True if API key is valid, otherwise False
        :rtype: bool
        :raises APIError: If there were any network- or server-related \
            issues while check the API key
        """
        try:
            self.metadata.categories()
            return True
        except UnauthorizedError:
            return False


class RealtoConnection:
    """
    The Time Series API client for customers subscribing via Realto.

    Wraps a Energy Quantified Time Series API on Realto. Handles validation,
    network errors, rate limiting, and parsing of the API responses.

    You must specify the **api_url**. There are

    Exactly one of **api_key** and **api_key_file** must be set at
    the same time. **api_key_file** shall be the file path to a file
    that contains just the API key (blank lines a stripped when read).

    :param api_url: The root URL for the API, such as
    :type api_url: string, optional
    :param api_key: The API key for your user account
    :type api_key: string, optional
    :param api_key_file: A file path to a file that contains the API key
    :type api_key_file: string, file, optional
    :param ssl_verify: Whether or not to verify the server certificate,\
                       defaults to True
    :type ssl_verify: bool, optional
    :param timeout: Maximum timeout per HTTP request, defaults to 15.0
    :type timeout: float, optional
    :param http_delay: The minimum number of seconds between the start of\
                       each HTTP request, defaults to 0.125 seconds (8 req/s)
    :type http_delay: float, optional

    **Basic usage:**

       >>> from energyquantified import RealtoConnection
       >>> eq = RealtoConnection(
       >>>     api_url=RealtoConnection.API_URL_GERMANY,
       >>>     api_key="aaaa-bbbb-cccc-dddd"
       >>> )
       >>> eq.metadata.curves(q="de wind power actual")

    """

    #: The base URL for the German data API on Realto.
    API_URL_GERMANY = 'https://api.realto.io/energyquantified-germany'

    #: The base URL for the French data API on Realto.
    API_URL_FRANCE = 'https://api.realto.io/energyquantified-france'

    #: The base URL for the Netherlands data API on Realto.
    API_URL_NETHERLANDS = 'https://api.realto.io/energyquantified-netherlands'

    #: The base URL for the UK data API on Realto.
    API_URL_UK = 'https://api.realto.io/energyquantified-greatbritain'

    #: The base URL for the Belgium data API on Realto.
    API_URL_BELGIUM = 'https://api.realto.io/energyquantified-belgium'

    def __init__(
            self,
            api_url=None,
            api_key=None,
            api_key_file=None,
            ssl_verify=True,
            timeout=15.0,
            http_delay=0.125
        ):
        # Simple validations
        assert api_url, "api_url is missing"
        assert api_key or api_key_file, "api_key is missing"
        if api_key:
            assert not api_key_file, (
                "Exactly one of 'api_key' and 'api_key_file' "
                "must be set at the same time, but both were given"
            )
        if api_key_file:
            assert not api_key, (
                "Exactly one of 'api_key' and 'api_key_file' "
                "must be set at the same time, but both were given"
            )
        assert timeout >= 0, "timeout must be larger than 0s"
        assert http_delay >= 0.0667, (
            "http_delay must be 0.0667s or slower (15 req/s)"
        )
        # Attributes
        self._api_url = api_url
        self._api_key = _find_api_key(api_key, api_key_file)
        # HTTP client
        self._session = Session(
            timeout=timeout,
            verify=ssl_verify,
            base_url=self._api_url,
            headers={
                "Ocp-Apim-Subscription-Key": self._api_key,
            },
            delay=http_delay
        )
        # --- Public members ---
        #: See :py:class:`energyquantified.api.MetadataAPI`. For metadata
        #: queries (such as curve search and place lookups).
        self.metadata = RealtoMetadataAPI(self)
        #: See :py:class:`energyquantified.api.TimeseriesAPI`. For loading
        #: time series data.
        self.timeseries = TimeseriesAPI(self)
        #: See :py:class:`energyquantified.api.InstancesAPI`. For loading
        #: time series instances.
        self.instances = InstancesAPI(self)
        #: See :py:class:`energyquantified.api.PeriodsAPI`. For
        #: loading period-based series.
        self.periods = PeriodsAPI(self)
        #: See :py:class:`energyquantified.api.PeriodInstancesAPI`. For
        #: loading instances of period-based series.
        self.period_instances = PeriodInstancesAPI(self)

    def is_api_key_valid(self):
        """
        Check if the supplied API key is valid/user account can sign in.

        Does so by trying to load the categories list. If it succeeds, then
        the user has supplied a valid API key.

        :return: True if API key is valid, otherwise False
        :rtype: bool
        :raises APIError: If there were any network- or server-related \
            issues while check the API key
        """
        try:
            self.metadata.curves(page_size=10)
            return True
        except UnauthorizedError:
            return False


def _find_api_key(api_key, api_key_file):
    # With the API key, it's easy
    if api_key:
        return api_key
    # Open the file and read the key
    try:
        with open(api_key_file, mode='r', encoding='utf-8') as f:
            api_key = f.read().strip()
            if not api_key:
                raise InitializationError(
                    f"Could not read API key from file '{api_key_file}' "
                    "because it was empty"
                )
            return api_key
    except IOError as e:
        raise InitializationError(
            f"Could not read API key from file '{api_key_file}'. "
            f"Reason: {e}"
        ) from e



from .api import (
    APIError,
    HTTPError,
    ValidationError,
    NotFoundError,
    UnauthorizedError,
    ForbiddenError,
    InternalServerError
)
from .client import InitializationError
from .page import PageError
from .parser import ParseException



from datetime import datetime

from .frequency import Frequency
from .utils import to_timezone


class Resolution:
    """
    A combination of a frequency and a time zone with utility methods
    for iteration and snapping datetimes to given frequency.

    :param frequency: The frequency for this resolution
    :type frequency: Frequency
    :param timezone: The time-zone for this resolution
    :type timezone: TzInfo (or similar pytz time-zone)
    """

    def __init__(self, frequency, timezone):
        assert isinstance(frequency, Frequency), "Invalid frequency"
        assert timezone, "Invalid or missing timezone"
        self.frequency = frequency
        self.timezone = timezone

    def __str__(self):
        return (
            "<Resolution: "
            f"frequency={self.frequency.tag}, timezone={self.timezone}"
            ">"
        )

    def __repr__(self):
        return str(self)

    def is_iterable(self):
        """
        Check whether or not this Resolution is iterable

        :return: True if this Resolution is iterable, otherwise False
        :rtype: bool
        """
        return self.frequency.is_iterable

    def to_dict(self):
        """
        Convert this resolution to a dict in the same format as data is
        returned from the Energy Quantified Time Series API.

        :return: A dict of this resolution in the same format as the API response
        :rtype: dict
        """
        return {
            "frequency": self.frequency.tag,
            "timezone": str(self.timezone),
        }

    def now(self):
        """
        Create a datetime, snapped to beginning of the current period of
        this resolution's frequency.

        :return: A date-time of now, snapped to beginning of the resolution's\
                 frequency
        :rtype: datetime
        """
        dt_with_tz = datetime.now(tz=self.timezone)
        return self.floor(dt_with_tz)

    def datetime(self, year=None, month=1, day=1, hour=0, minute=0, second=0,
                 millis=0):
        """
        Create a datetime. Only year needs to be specified. The remaining
        fields defaults to the lowest value. The datetime will have the same
        timezone as this resolution and will be snapped to the frequency.

        :param year: The year
        :type year: int, required
        :param month: The month (1-12), defaults to 1
        :type month: int, optional
        :param day: The day-of-month (1-31), defaults to 1
        :type day: int, optional
        :param hour: The hour of the day (0-23), defaults to 0
        :type hour: int, optional
        :param minute: The minute (0-59), defaults to 0
        :type minute: int, optional
        :param second: The second (0-59), defaults to 0
        :type second: int, optional
        :param millis: The millisecond (0-999), defaults to 0
        :type millis: int, optional
        :return: A new date-time, snapped down to the frequency
        :rtype: datetime
        """
        assert year and 1900 <= year <= 2100, "Must specify year (1900-2100)"
        dt = datetime(year, month, day, hour, minute, second, millis * 1000)
        dt_with_tz = self.timezone.localize(dt)
        return self.floor(dt_with_tz)

    def matches(self, datetime_obj):
        """
        Check if a datetime obj has the same time zone as this resolution
        and that the datetime aligns with the frequency of this resolution.

        :param datetime_obj: A date-time
        :type datetime_obj: datetime
        :return: True if it falls into this frequency, otherwise False
        :rtype: bool
        """
        if self.frequency <= Frequency.P1D:
            return self.frequency.matches(datetime_obj)
        else:
            return (
                datetime_obj.tzinfo.zone == self.timezone.zone
                and self.frequency.matches(datetime_obj)
            )

    def __rshift__(self, datetime_obj):
        """
        Move a datetime one step forward in time. Usage:

           >>> datetime_obj = resolution >> datetime_obj

        :param datetime_obj: A date-time
        :type datetime_obj: datetime
        :return: The supplied date-time shifted one step forward in time
        :rtype: datetime
        """
        assert self.is_iterable(), (
            "Operator >> requires an iterable Resolution"
        )
        # self >> datetime_obj to move one step forward
        return self.shift(datetime_obj, 1)

    def __lshift__(self, datetime_obj):
        """
        Move a datetime one step backward in time. Usage:

           >>> datetime_obj = resolution << datetime_obj

        :param datetime_obj: A date-time
        :type datetime_obj: datetime
        :return: The supplied date-time shifted one step backward in time
        :rtype: datetime
        """
        assert self.is_iterable(), (
            "Operator << requires an iterable Resolution"
        )
        # self << datetime_obj to move one step backward
        return self.shift(datetime_obj, -1)

    def shift(self, datetime_obj, steps):
        """
        Shift a date-time by N steps in this resolution's frequency. 0 steps
        means no change. 1 step means one tick forward, -1 step means one
        tick backwards.

        :param datetime_obj: A date-time
        :type datetime_obj: datetime
        :param steps: Number of steps forward (positive) or backwards (negative)
        :type steps: int
        :return: A shifted date-time
        :rtype: datetime
        """
        assert self.is_iterable(), (
            "Resolution.shift() requires an iterable Resolution"
        )
        self._assert_valid(datetime_obj)
        return self.frequency.shift(datetime_obj, steps)

    def floor(self, datetime_obj):
        """
        Snap a date or datetime down to the nearest instant in this
        resolution's frequency.

        :param datetime_obj: A date-time
        :type datetime_obj: datetime
        :return: A date-time snapped down to the nearest instant in this\
                 resolution's frequency.
        :rtype: datetime
        """
        # Make sure it is in the correct time zone and truncate
        dt = to_timezone(datetime_obj, tz=self.timezone)
        return self.frequency.truncate(dt)

    def ceil(self, datetime_obj):
        """
        Snap a date or datetime up to the nearest instant in this
        resolution's frequency.

        :param datetime_obj: A date-time
        :type datetime_obj: datetime
        :return: A date-time snapped up to the nearest instant in this\
                 resolution's frequency.
        :rtype: datetime
        """
        # Floor
        truncated = self.floor(datetime_obj)
        # Shift one step if not already in resolution
        return self >> truncated if truncated != datetime_obj else truncated

    def enumerate(self, begin=None, end=None):
        """
        Returns a generator over all datetimes between `begin` and `end` in
        this resolution. Example:

        >>> resolution = Resolution(Frequency.P1D, CET)
        >>> begin = resolution.datetime(2020, 5, 16)
        >>> end = resolution.datetime(2020, 5, 26)
        >>> for i in resolution.enumerate(begin, end):
                print(i)

        :param begin: The begin date-time (inclusive)
        :type begin: datetime, required
        :param end: The end date-time (exclusive)
        :type end: datetime, required
        :yield: A generator of date-times between `begin` and `end` in\
                this resolution
        :rtype: datetime
        """
        assert self.is_iterable(), (
            "Resolution.enumerate() requires an iterable Resolution"
        )
        assert begin, "begin is not set"
        assert end, "end is not set"
        self._assert_valid(begin)
        d = begin
        while d < end:
            yield d
            d = self >> d

    def _assert_valid(self, datetime_obj):
        assert self.matches(datetime_obj), (
            "datetime_obj %s does not match %s" % (datetime_obj, self)
        )
